# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MAPPO builder and config."""
import dataclasses
from typing import Dict, Iterator, List, Optional, Type, Union

import reverb
import sonnet as snt
import tensorflow as tf
from acme import types as acme_types
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils
from acme.utils import counting

from mava import adders, core, specs, types
from mava.adders import reverb as reverb_adders
from mava.systems.builders import SystemBuilder
from mava.systems.tf.mappo import execution, training
from mava.wrappers import DetailedTrainerStatistics, NetworkStatisticsActorCritic


@dataclasses.dataclass
class MAPPOConfig:
    """Configuration options for the MAPPO system
    Args:
        environment_spec: environment specs.
        sequence_length: length of the sequences in the queue.
        sequence_period: amount of overlap between sequences added to the queue.
        discount: discount factor (Always between 0 and 1).
        lambda_gae: scalar determining the mix of bootstrapping
            vs further accumulation of multi-step returns at each timestep.
            See `High-Dimensional Continuous Control Using Generalized
            Advantage Estimation` for more information.
        clipping_epsilon: Hyper-parameter for clipping in the policy
            objective. Roughly: how far can the new policy go from
            the old policy while still profiting? The new policy can
            still go farther than the clip_ratio says, but it doesn’t
            help on the objective anymore.
        max_queue_size: maximum number of items in the queue.
        batch_size: size of the mini-batches.
        critic_optimizer: optimizer for the critic.
        policy_optimizer: optimizer for the policy.
        entropy_cost: contribution of entropy regularization to the total loss.
        baseline_cost: contribution of the value loss to the total loss.
        max_abs_reward: max reward. If not None, the reward on which the agent
            is trained will be clipped between -max_abs_reward and max_abs_reward.
        max_gradient_norm: gradient normalization.
        replay_table_name: string indicating what name to give the replay table.
    """

    environment_spec: specs.EnvironmentSpec
    policy_optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]]
    critic_optimizer: snt.Optimizer
    sequence_length: int = 10
    sequence_period: int = 5
    shared_weights: bool = False
    discount: float = 0.99
    lambda_gae: float = 0.95
    max_queue_size: int = 100_000
    executor_variable_update_period: int = 100
    batch_size: int = 32
    entropy_cost: float = 0.01
    baseline_cost: float = 0.5
    clipping_epsilon: float = 0.1
    max_gradient_norm: Optional[float] = None
    checkpoint: bool = True
    checkpoint_subpath: str = "~/mava/"
    replay_table_name: str = reverb_adders.DEFAULT_PRIORITY_TABLE


class MAPPOBuilder(SystemBuilder):

    """Builder for MAPPO which constructs individual components of the system."""

    """Defines an interface for defining the components of a MARL system.
      Implementations of this interface contain a complete specification of a
      concrete MARL system. An instance of this class can be used to build a
      MARL system which interacts with the environment either locally or in a
      distributed setup.
      """

    def __init__(
        self,
        config: MAPPOConfig,
        trainer_fn: Type[training.MAPPOTrainer] = training.MAPPOTrainer,
        executor_fn: Type[core.Executor] = execution.MAPPOFeedForwardExecutor,
    ):
        """Args:
        config: Configuration options for the MAPPO system.
        policy_networks: ...
        trainer_fn: ...
        """

        self._config = config
        self._agents = self._config.environment_spec.get_agent_ids()
        self._agent_types = self._config.environment_spec.get_agent_types()
        self._trainer_fn = trainer_fn
        self._executor_fn = executor_fn

    def make_replay_tables(
        self,
        environment_spec: specs.MAEnvironmentSpec,
    ) -> List[reverb.Table]:
        """Create tables to insert data into."""
        agent_specs = environment_spec.get_agent_specs()
        extras_spec: Dict[str, Dict[str, acme_types.NestedArray]] = {"log_probs": {}}

        for agent, spec in agent_specs.items():
            # Make dummy log_probs
            extras_spec["log_probs"][agent] = tf.ones(shape=(1,), dtype=tf.float32)

        # Squeeze the batch dim.
        extras_spec = tf2_utils.squeeze_batch_dim(extras_spec)

        replay_table = reverb.Table.queue(
            name=self._config.replay_table_name,
            max_size=self._config.max_queue_size,
            signature=reverb_adders.ParallelSequenceAdder.signature(
                environment_spec, extras_spec=extras_spec
            ),
        )

        return [replay_table]

    def make_dataset_iterator(
        self,
        replay_client: reverb.Client,
    ) -> Iterator[reverb.ReplaySample]:
        """Create a dataset iterator to use for learning/updating the system.
        Args:
          replay_client: ...
        """
        # Create tensorflow dataset to interface with reverb
        dataset = reverb.ReplayDataset.from_table_signature(
            server_address=replay_client.server_address,
            table=self._config.replay_table_name,
            max_in_flight_samples_per_worker=1,
            sequence_length=self._config.sequence_length,
            emit_timesteps=False,
        )

        # Batching
        dataset = dataset.batch(self._config.batch_size, drop_remainder=True)

        return iter(dataset)

    def make_adder(
        self,
        replay_client: reverb.Client,
    ) -> Optional[adders.ParallelAdder]:
        """Create an adder which records data generated by the executor/environment.
        Args:
          replay_client: a Reverb client which points to the replay server.
        """
        return reverb_adders.ParallelSequenceAdder(
            client=replay_client,
            period=self._config.sequence_period,
            sequence_length=self._config.sequence_length,
            use_next_extras=False,
        )

    def make_executor(
        self,
        policy_networks: Dict[str, snt.Module],
        adder: Optional[adders.ParallelAdder] = None,
        variable_source: Optional[core.VariableSource] = None,
    ) -> core.Executor:

        """Create an executor instance.
        Args:
            policy_networks: a struct of instance of all the different networks.
            adder: how data is recorded (e.g. added to replay).
            variable_source: a source providing the necessary executor parameters.
        """

        variable_client = None
        if variable_source:
            # Create policy variables.
            variables = {
                network_key: policy_networks[network_key].variables
                for network_key in policy_networks
            }

            # Get new policy variables.
            # TODO Should the variable update period be in the MAPPO config?
            variable_client = variable_utils.VariableClient(
                client=variable_source,
                variables={"policy": variables},
                update_period=self._config.executor_variable_update_period,
            )

            # Make sure not to use a random policy after checkpoint restoration by
            # assigning variables before running the environment loop.
            variable_client.update_and_wait()

        # Create the executor which defines how agents take actions.
        return self._executor_fn(
            policy_networks=policy_networks,
            shared_weights=self._config.shared_weights,
            variable_client=variable_client,
            adder=adder,
        )

    def make_trainer(
        self,
        networks: Dict[str, Dict[str, snt.Module]],
        dataset: Iterator[reverb.ReplaySample],
        replay_client: Optional[reverb.Client] = None,
        counter: Optional[counting.Counter] = None,
        logger: Optional[types.NestedLogger] = None,
    ) -> core.Trainer:
        """Creates an instance of the trainer.
        Args:
            networks: a struct describing the networks needed by the trainer; this can
                be specific to the trainer in question.
            dataset: iterator over samples from replay.
            replay_client: client which allows communication with replay, e.g. in
                order to update priorities.
            counter: a Counter which allows for recording of counts (trainer steps,
                executor steps, etc.) distributed throughout the system.
            logger: Logger object for logging metadata.
            checkpoint: bool controlling whether the trainer checkpoints itself.
        """
        agents = self._agents
        agent_types = self._agent_types
        shared_weights = self._config.shared_weights

        observation_networks = networks["observations"]
        policy_networks = networks["policies"]
        critic_networks = networks["critics"]

        # The learner updates the parameters (and initializes them).
        trainer = self._trainer_fn(
            agents=agents,
            agent_types=agent_types,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            dataset=dataset,
            shared_weights=shared_weights,
            critic_optimizer=self._config.critic_optimizer,
            policy_optimizer=self._config.policy_optimizer,
            discount=self._config.discount,
            lambda_gae=self._config.lambda_gae,
            entropy_cost=self._config.entropy_cost,
            baseline_cost=self._config.baseline_cost,
            clipping_epsilon=self._config.clipping_epsilon,
            max_gradient_norm=self._config.max_gradient_norm,
            counter=counter,
            logger=logger,
            checkpoint=self._config.checkpoint,
            checkpoint_subpath=self._config.checkpoint_subpath,
        )

        trainer = NetworkStatisticsActorCritic(trainer)  # type: ignore

        trainer = DetailedTrainerStatistics(  # type: ignore
            trainer, metrics=["policy_loss", "critic_loss"]
        )

        return trainer



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Adders that use Reverb (github.com/deepmind/reverb) as a backend."""

import abc
import collections
from typing import (
    Callable,
    Deque,
    Dict,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import dm_env
import numpy as np
import reverb
import tensorflow as tf
import tree
from acme import specs as acme_specs
from acme import types

from mava import specs as mava_specs
from mava.adders import base

DEFAULT_PRIORITY_TABLE = "priority_table"


class Step(NamedTuple):
    """Step class used internally for reverb adders."""

    observations: Dict[str, types.NestedArray]
    actions: Dict[str, types.NestedArray]
    rewards: Dict[str, types.NestedArray]
    discounts: Dict[str, types.NestedArray]
    start_of_episode: Union[bool, acme_specs.Array, tf.Tensor, Tuple[()]]
    extras: Dict[str, types.NestedArray]


class PriorityFnInput(NamedTuple):
    """The input to a priority function consisting of stacked steps."""

    observations: Dict[str, types.NestedArray]
    actions: Dict[str, types.NestedArray]
    rewards: Dict[str, types.NestedArray]
    discounts: Dict[str, types.NestedArray]
    start_of_episode: types.NestedArray
    extras: Dict[str, types.NestedArray]


# Define the type of a priority function and the mapping from table to function.
PriorityFn = Callable[["PriorityFnInput"], float]
PriorityFnMapping = Mapping[str, Optional[PriorityFn]]


def spec_like_to_tensor_spec(
    paths: Iterable[str], spec: acme_specs.Array
) -> tf.TypeSpec:
    return tf.TensorSpec.from_spec(spec, name="/".join(str(p) for p in paths))


class ReverbParallelAdder(base.ParallelAdder):
    """Base class for Reverb adders."""

    def __init__(
        self,
        client: reverb.Client,
        buffer_size: int,
        max_sequence_length: int,
        delta_encoded: bool = False,
        chunk_length: Optional[int] = None,
        priority_fns: Optional[PriorityFnMapping] = None,
        max_in_flight_items: Optional[int] = 25,
        use_next_extras: bool = True,
    ):
        """Initialize a ReverbAdder instance.
        Args:
          client: A client to the Reverb backend.
          buffer_size: Number of steps to retain in memory.
          max_sequence_length: The maximum length of sequences (corresponding to the
            number of observations) that can be added to replay.
          delta_encoded: If `True` (False by default) enables delta encoding, see
            `Client` for more information.
          chunk_length: Number of timesteps grouped together before delta encoding
            and compression. See `Client` for more information.
          priority_fns: A mapping from table names to priority functions; if
            omitted, all transitions/steps/sequences are given uniform priorities
            (1.0) and placed in DEFAULT_PRIORITY_TABLE.
          max_in_flight_items: The maximum number of items allowed to be "in flight"
            at the same time. See `reverb.Writer.writer` for more info.
        """
        if priority_fns:
            priority_fns = dict(priority_fns)
        else:
            priority_fns = {DEFAULT_PRIORITY_TABLE: lambda x: 1.0}

        self._client = client
        self._priority_fns = priority_fns
        self._max_sequence_length = max_sequence_length
        self._delta_encoded = delta_encoded
        self._chunk_length = chunk_length
        self._max_in_flight_items = max_in_flight_items
        self._use_next_extras = use_next_extras

        # This is exposed as the _writer property in such a way that it will create
        # a new writer automatically whenever the internal __writer is None. Users
        # should ONLY ever interact with self._writer.
        self.__writer = None

        # The state of the adder is captured by a buffer of `buffer_size` steps
        # (generally SAR tuples) and one additional dangling observation.
        self._buffer: Deque = collections.deque(maxlen=buffer_size)
        self._next_extras: Union[None, Dict[str, types.NestedArray]] = None
        self._next_observations = None
        self._start_of_episode = False

    def __del__(self) -> None:
        if self.__writer is not None:
            # Explicitly close the writer with no retry on server unavailable.
            # This is to avoid hang on closing if the server has already terminated.
            self.__writer.close(retry_on_unavailable=False)

    @property
    def _writer(self) -> reverb.Writer:
        if self.__writer is None:
            self.__writer = self._client.writer(
                self._max_sequence_length,
                delta_encoded=self._delta_encoded,
                chunk_length=self._chunk_length,
                max_in_flight_items=self._max_in_flight_items,
            )
        return self.__writer

    def add_priority_table(
        self, table_name: str, priority_fn: Optional[PriorityFn]
    ) -> None:
        if table_name in self._priority_fns:
            raise ValueError(
                "A priority function already exists for {}.".format(table_name)
            )
        self._priority_fns[table_name] = priority_fn

    def reset(self) -> None:
        """Resets the adder's buffer."""
        if self.__writer:
            self._writer.close()
            self.__writer = None
        self._buffer.clear()
        self._next_observations = None

    def add_first(
        self, timestep: dm_env.TimeStep, extras: Dict[str, types.NestedArray] = {}
    ) -> None:
        """Record the first observation of a trajectory."""
        if not timestep.first():
            raise ValueError(
                "adder.add_first with an initial timestep (i.e. one for "
                "which timestep.first() is True"
            )

        if self._next_observations is not None:
            raise ValueError(
                "adder.reset must be called before adder.add_first "
                "(called automatically if `next_timestep.last()` is "
                "true when `add` is called)."
            )

        # Record the next observation.
        self._next_observations = timestep.observation
        self._start_of_episode = True

        if self._use_next_extras:
            self._next_extras = extras

    def add(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        next_extras: Dict[str, types.NestedArray] = {},
    ) -> None:
        """Record an action and the following timestep."""
        if self._next_observations is None:
            raise ValueError("adder.add_first must be called before adder.add.")

        discount = next_timestep.discount
        if next_timestep.last():
            # print("LAST")
            # Terminal timesteps created by dm_env.termination() will have a scalar
            # discount of 0.0. This may not match the array shape / nested structure
            # of the previous timesteps' discounts. The below will match
            # next_timestep.discount's shape/structure to that of
            # self._buffer[-1].discount.
            if self._buffer and not tree.is_nested(next_timestep.discount):
                discount = tree.map_structure(
                    lambda d: np.broadcast_to(next_timestep.discount, np.shape(d)),
                    self._buffer[-1].discount,
                )

        self._buffer.append(
            Step(
                observations=self._next_observations,
                actions=actions,
                rewards=next_timestep.reward,
                discounts=discount,
                start_of_episode=self._start_of_episode,
                extras=self._next_extras if self._use_next_extras else next_extras,
            )
        )

        # Write the last "dangling" observation.
        if next_timestep.last():
            self._start_of_episode = False
            self._write()
            self._write_last()
            self.reset()
        else:
            # Record the next observation and write.
            # Possibly store next_extras
            if self._use_next_extras:
                self._next_extras = next_extras
            self._next_observations = next_timestep.observation
            self._start_of_episode = False
            self._write()

    @abc.abstractmethod
    def signature(
        cls,
        environment_spec: mava_specs.MAEnvironmentSpec,
        extras_spec: tf.TypeSpec,
    ) -> tf.TypeSpec:
        """This is a helper method for generating signatures for Reverb tables.
        Signatures are useful for validating data types and shapes, see Reverb's
        documentation for details on how they are used.
        Args:
          environment_spec: A `specs.EnvironmentSpec` whose fields are nested
            structures with leaf nodes that have `.shape` and `.dtype` attributes.
            This should come from the environment that will be used to generate
            the data inserted into the Reverb table.
          core_state_spec: A nested structure with leaf nodes that have `.shape` and
            `.dtype` attributes. The structure (and shapes/dtypes) of this must
            be the same as the `core_state` passed into `ReverbAdder.add`.
        Returns:
          A `Step` whose leaf nodes are `tf.TensorSpec` objects.
        """

    @abc.abstractmethod
    def _write(self) -> None:
        """Write data to replay from the buffer."""

    @abc.abstractmethod
    def _write_last(self) -> None:
        """Write data to replay from the buffer."""



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (StJohn): complete class for transformed mixing
# https://github.com/oxwhirl/wqmix/blob/master/src/modules/mixers/qtran.py
# https://github.com/starry-sky6688/StarCraft/blob/master/network/qtran_net.py


"""Mixing for multi-agent RL systems"""

from typing import Optional, Tuple

import sonnet as snt
import tensorflow as tf


class QTranBase(snt.Module):
    """Multi-agent mixing architecture."""

    def __init__(
        self,
        n_agents: int,
        n_actions: int,
        state_shape: Tuple,
        qtran_hidden_dim: int,
        rnn_hidden_dim: int,
        qtran_arch: str = "qtran_paper",
        larger_network: bool = True,
    ) -> None:
        """Initializes the mixer."""
        super(QTranBase, self).__init__()

        self._n_agents = n_agents
        self._n_actions = n_actions
        self._state_dim = int(tf.prod(state_shape))
        self._qtran_arch = qtran_arch
        self._qtran_hidden_dim = qtran_hidden_dim
        self._larger_network = larger_network
        self._rnn_hidden_dim = rnn_hidden_dim

        self._q_input_size = 0
        if self._qtran_arch == "coma_critic":
            self._q_input_size = self._state_dim + (self._n_agents * self._n_actions)
        elif self._qtran_arch == "qtran_paper":
            self._q_input_size = (
                self._state_dim + self._rnn_hidden_dim + self._n_actions
            )
        else:
            raise Exception(
                "{} is not a valid QTran architecture".format(self._qtran_arch)
            )

        # Define Q and V networks
        if self._larger_network:  # 3 hidden layers
            self._Q = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
            self._V = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
        else:  # 2 hidden layers
            self._Q = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
            self._V = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )

        # Action encoding
        ae_input = self._rnn_hidden_dim + self._n_actions
        self._action_encoding = snt.Sequential(
            [
                snt.Flatten(preserve_dims=1),
                snt.Linear(ae_input),
                snt.relu(),
                snt.Linear(ae_input),
            ]
        )

    def __call__(
        self,
        batch: tf.Tensor,  # Convert this to whatever reverb uses.
        hidden_states: tf.Tensor,
        actions: Optional[tf.Tensor] = None,
    ) -> tf.Tensor:
        """Perform mixing logic"""
        bs = batch.batch_size
        ts = batch.max_seq_length
        states = tf.reshape(batch["state"], shape=(bs * ts, self._state_dim))

        if actions is None:
            # Use the actions taken by the agents
            actions = tf.reshape(
                batch["actions_onehot"],
                shape=(bs * ts, self._n_agents * self._n_actions),
            )
        else:
            # It will arrive as (bs, ts, agents, actions), we need to reshape it
            actions = tf.reshape(
                actions, shape=(bs * ts, self._n_agents * self._n_actions)
            )

        if self._qtran_arch == "coma_critic":
            inputs = tf.concat((states, actions), axis=1)
        elif self._qtran_arch == "qtran_paper":
            hidden_states = tf.reshape(
                hidden_states, shape=(bs * ts, self._n_agents, -1)
            )
            agent_state_action_input = tf.concat((hidden_states, actions), axis=2)

            agent_state_action_encoding = self._action_encoding(
                tf.reshape(
                    agent_state_action_input, shape=(bs * ts * self._n_agents, -1)
                )
            )
            agent_state_action_encoding = tf.reshape(
                agent_state_action_encoding, shape=(bs * ts, self._n_agents, -1)
            )
            agent_state_action_encoding = tf.math.reduce_sum(
                agent_state_action_encoding, axis=1
            )  # Sum across agents

            inputs = tf.concat((states, agent_state_action_encoding), axis=1)

        q_outputs = self._Q(inputs)
        states = tf.reshape(batch["state"], shape=(bs * ts, self._state_dim))
        v_outputs = self._V(states)

        return q_outputs, v_outputs


class QTranAlt(snt.Module):
    def __init__(
        self,
        n_agents: int,
        n_actions: int,
        state_shape: Tuple,
        qtran_hidden_dim: int,
        rnn_hidden_dim: int,
        qtran_arch: str = "qtran_paper",
        larger_network: bool = True,
    ) -> None:
        """Initializes the mixer."""
        super(QTranAlt, self).__init__()

        self._n_agents = n_agents
        self._n_actions = n_actions
        self._state_dim = int(tf.prod(state_shape))
        self._qtran_arch = qtran_arch
        self._qtran_hidden_dim = qtran_hidden_dim
        self._larger_network = larger_network
        self._rnn_hidden_dim = rnn_hidden_dim

        # Q(s,-,u-i)
        # Q takes [state, u-i, i] as input
        self._q_input_size = self._state_dim + (self._n_agents * self._n_actions)

        # Define Q and V networks
        if self._larger_network:  # 3 hidden layers
            self._Q = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
            self._V = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
        else:  # 2 hidden layers
            self._Q = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )
            self._V = snt.Sequential(
                [
                    snt.Flatten(preserve_dims=1),
                    snt.Linear(self._qtran_hidden_dim),
                    tf.nn.relu(),
                    snt.Linear(1),
                ]
            )

    def __call__(
        self,
        batch: tf.Tensor,  # Convert this to whatever reverb uses.
        masked_actions: Optional[tf.Tensor] = None,
    ) -> tf.Tensor:
        """Perform mixing logic"""
        bs = batch.batch_size
        ts = batch.max_seq_length

        # NOTE (St John) Check this carefully. I'm not sure what exactly batch will
        # contain. My use of tf.tile() here is assuming tf is a tensor. Behaviour is
        # unchecked at this stage.

        # Repeat each state n_agents times
        repeated_states = tf.tile(batch["state"], multiples=[1, 1, self._n_agents])
        repeated_states = tf.reshape(repeated_states, shape=(-1, self._state_dim))

        if masked_actions is None:
            # Check what shape the incoming actions have when debugging.
            actions = tf.tile(
                batch["actions_onehot"], multiples=[1, 1, self._n_agents, 1]
            )

            agent_mask = 1 - tf.eye(self._n_agents)  # [n_agents,n_agents]
            agent_mask = tf.reshape(agent_mask, shape=(-1, 1))  # [n_agents^2,1]
            agent_mask = tf.tile(
                agent_mask, [1, self._n_actions]
            )  # [n_agents^2,n_actions]
            # Add extra dimensions for multiplication coming up
            # [1,1,n_agents^2,n_actions]
            agent_mask = tf.expand_dims(tf.expand_dims(agent_mask, 0), 0)

            masked_actions = actions * agent_mask
            masked_actions = tf.reshape(
                masked_actions, shape=(-1, self._n_agents * self._n_actions)
            )

        agent_ids = tf.eye(self._n_agents)  # [n_agents,n_agents]
        agent_ids = tf.expand_dims(
            tf.expand_dims(agent_ids, 0), 0
        )  # [1,1,n_agents,n_agents]
        agent_ids = tf.tile(agent_ids, [bs, ts, 1, 1])  # [bs,ts,n_agents,n_agents]
        agent_ids = tf.reshape(
            agent_ids, shape=(-1, self._n_agents)
        )  # [bs*ts*n_agents,n_agents]

        inputs = tf.concat([repeated_states, masked_actions, agent_ids], axis=1)

        q_outputs = self._Q(inputs)

        states = tf.tile(batch["state"], [1, 1, self._n_agents])
        states = tf.reshape(states, shape=(-1, self._state_dim))

        v_outputs = self._V(states)

        return q_outputs, v_outputs



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import sonnet as snt
import tensorflow as tf
from tensorflow import Tensor


class HyperNetwork(snt.Module):
    def __init__(
        self,
        agent_networks: Dict[str, snt.Module],
        qmix_hidden_dim: int,  # qmix_hidden_dim
        num_hypernet_layers: int = 2,
        hypernet_hidden_dim: int = 0,  # qmix_hidden_dim
    ):
        """Initializes the mixer.
        Args:
            agent_networks: Networks which produce outputs for mixing network.
            qmix_hidden_dim: Mixing layers hidden dimensions.
                i.e. What size the mixing network takes as input.
            num_hypernet_layers: Number of hypernetwork layers. Currently 1 or 2.
            hypernet_hidden_dim: The number of nodes in the hypernetwork hidden
                layer. Relevant for num_hypernet_layers > 1.
        """
        super(HyperNetwork, self).__init__()
        self._agent_networks = agent_networks
        self._qmix_hidden_dim = qmix_hidden_dim
        self._num_hypernet_layers = num_hypernet_layers

        # Let the user define the hidden dim but default it to qmix_hidden_dim.
        if hypernet_hidden_dim == 0:
            self._hypernet_hidden_dim = qmix_hidden_dim
        else:
            self._hypernet_hidden_dim = hypernet_hidden_dim

        # Calculate other necessary variables
        self._n_agents = len(agent_networks["values"])

        # Set up hypernetwork configuration
        if self._num_hypernet_layers == 1:
            self.hyper_w1 = snt.nets.MLP(
                output_sizes=[self._qmix_hidden_dim * self._n_agents]
            )
            self.hyper_w2 = snt.nets.MLP(output_sizes=[self._qmix_hidden_dim])

        # Default
        elif self._num_hypernet_layers == 2:
            self.hyper_w1 = snt.nets.MLP(
                output_sizes=[
                    self._hypernet_hidden_dim,
                    self._qmix_hidden_dim * self._n_agents,
                ]
            )
            self.hyper_w2 = snt.nets.MLP(
                output_sizes=[self._hypernet_hidden_dim, self._qmix_hidden_dim]
            )

        # State dependent bias for hidden layer
        self.hyper_b1 = snt.nets.MLP(output_sizes=[self._qmix_hidden_dim])
        self.hyper_b2 = snt.nets.MLP(output_sizes=[self._qmix_hidden_dim, 1])

    def __call__(self, states: Tensor) -> Dict[str, float]:  # [batch_size=B, state_dim]
        w1 = tf.abs(
            self.hyper_w1(states)
        )  # [B, qmix_hidden_dim] = [B, qmix_hidden_dim]
        w1 = tf.reshape(
            w1,
            (-1, self._n_agents, self._qmix_hidden_dim),
        )  # [B, n_agents, qmix_hidden_dim]

        b1 = self.hyper_b1(states)  # [B, qmix_hidden_dim] = [B, qmix_hidden_dim]
        b1 = tf.reshape(b1, [-1, 1, self._qmix_hidden_dim])  # [B, 1, qmix_hidden_dim]

        w2 = tf.abs(self.hyper_w2(states))
        w2 = tf.reshape(
            w2, shape=(-1, self._qmix_hidden_dim, 1)
        )  # [B, qmix_hidden_dim, 1]

        b2 = self.hyper_b2(states)  # [B, 1]
        b2 = tf.reshape(b2, shape=(-1, 1, 1))  # [B, 1, 1]

        hyperparams = {}
        hyperparams["w1"] = w1  # [B, n_agents, qmix_hidden_dim]
        hyperparams["b1"] = b1  # [B, 1, qmix_hidden_dim]
        hyperparams["w2"] = w2  # [B, qmix_hidden_dim]
        hyperparams["b2"] = b2  # [B, 1]

        return hyperparams



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore
class GameObject:
    """
    Root class for all percievable objects in the world model.
    """

    def __init__(self, distance, direction):
        """
        All objects have a distance and direction to the player, at a minimum.
        """

        self.distance = distance
        self.direction = direction


class Line(GameObject):
    """
    Represents a line on the soccer field.
    """

    def __init__(self, distance, direction, line_id):
        self.line_id = line_id

        GameObject.__init__(self, distance, direction)


class Goal(GameObject):
    """
    Represents a goal object on the field.
    """

    def __init__(self, distance, direction, goal_id):
        self.goal_id = goal_id

        GameObject.__init__(self, distance, direction)


class Flag(GameObject):
    """
    A flag on the field.  Can be used by the agent to determine its position.
    """

    # a dictionary mapping all flag_ids to their on-field (x, y) coordinates
    # TODO: Add real values form code here and do not use estimates.
    # TODO: these are educated guesses based on Figure 4.2 in the documentation.
    #       where would one find the actual coordinates, besides in the server
    #       code?
    # Location:
    # https://github.com/rcsoccersim/rcssserver/blob/master/src/landmarkreader.cpp
    # https://rcsoccersim.github.io/manual/soccerserver.html Fig 4.2

    pitch_x = 52.5
    pitch_y = 34.0
    outside_add = 5
    goal_y = 7.01
    pen_x = 36.0
    pen_y = 20.16

    out_x = pitch_x + outside_add
    out_y = pitch_y + outside_add

    FLAG_COORDS = {
        # perimiter flags
        # center flag
        "c": (0, 0),
        "ct": (0, pitch_y),
        "cb": (0, -pitch_y),
        # field boundary flags (on boundary lines)
        "r": (pitch_x, 0),
        "rt": (pitch_x, pitch_y),
        "rb": (pitch_x, -pitch_y),
        "l": (-pitch_x, 0),
        "lt": (-pitch_x, pitch_y),
        "lb": (-pitch_x, -pitch_y),
        # goal flags ('t' and 'b' flags can change based on server parameter
        # 'goal_width', but we leave their coords as the default values.
        # TODO: make goal flag coords dynamic based on server_params
        "grb": (pitch_x, -goal_y),
        "gr": (pitch_x, 0),
        "grt": (pitch_x, goal_y),
        "glb": (-pitch_x, -goal_y),
        "gl": (-pitch_x, 0),
        "glt": (-pitch_x, goal_y),
        # penalty flags
        "prb": (pen_x, -pen_y),
        "prc": (pen_x, 0),
        "prt": (pen_x, pen_y),
        "plb": (-pen_x, -pen_y),
        "plc": (-pen_x, 0),
        "plt": (-pen_x, pen_y),
        "t0": (0, out_y),
        "tr10": (10, out_y),
        "tr20": (20, out_y),
        "tr30": (30, out_y),
        "tr40": (40, out_y),
        "tr50": (50, out_y),
        "tl10": (-10, out_y),
        "tl20": (-20, out_y),
        "tl30": (-30, out_y),
        "tl40": (-40, out_y),
        "tl50": (-50, out_y),
        "b0": (0, -out_y),
        "br10": (10, -out_y),
        "br20": (20, -out_y),
        "br30": (30, -out_y),
        "br40": (40, -out_y),
        "br50": (50, -out_y),
        "bl10": (-10, -out_y),
        "bl20": (-20, -out_y),
        "bl30": (-30, -out_y),
        "bl40": (-40, -out_y),
        "bl50": (-50, -out_y),
        "r0": (out_x, 0),
        "rt10": (out_x, 10),
        "rt20": (out_x, 20),
        "rt30": (out_x, 30),
        "rb10": (out_x, -10),
        "rb20": (out_x, -20),
        "rb30": (out_x, -30),
        "l0": (-out_x, 0),
        "lt10": (-out_x, 10),
        "lt20": (-out_x, 20),
        "lt30": (-out_x, 30),
        "lb10": (-out_x, -10),
        "lb20": (-out_x, -20),
        "lb30": (-out_x, -30),
    }

    def __init__(self, distance, direction, flag_id):
        """
        Adds a flag id for this field object.  Every flag has a unique id.
        """

        self.flag_id = flag_id

        GameObject.__init__(self, distance, direction)


class MobileObject(GameObject):
    """
    Represents objects that can move.
    """

    def __init__(self, distance, direction, dist_change, dir_change, speed):
        """
        Adds variables for distance and direction deltas.
        """

        self.dist_change = dist_change
        self.dir_change = dir_change
        self.speed = speed

        GameObject.__init__(self, distance, direction)


class Ball(MobileObject):
    """
    A spcial instance of a mobile object representing the soccer ball.
    """

    def __init__(self, distance, direction, dist_change, dir_change, speed):
        MobileObject.__init__(self, distance, direction, dist_change, dir_change, speed)


class Player(MobileObject):
    """
    Represents a friendly or enemy player in the game.
    """

    def __init__(
        self,
        distance,
        direction,
        dist_change,
        dir_change,
        speed,
        team,
        side,
        uniform_number,
        body_direction,
        neck_direction,
    ):
        """
        Adds player-specific information to a mobile object.
        """

        self.team = team
        self.side = side
        self.uniform_number = uniform_number
        self.body_direction = body_direction
        self.neck_direction = neck_direction

        MobileObject.__init__(self, distance, direction, dist_change, dir_change, speed)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Commonly used centralised architectures for multi-agent RL systems"""

import copy
from typing import Dict, Tuple

import sonnet as snt
import tensorflow as tf
from acme import specs as acme_specs

from mava import specs as mava_specs
from mava.components.tf.architectures.decentralised import (
    DecentralisedPolicyActor,
    DecentralisedQValueActorCritic,
)


class StateBasedPolicyActor(DecentralisedPolicyActor):
    """Multi-agent actor architecture using
    environment state information."""

    def __init__(
        self,
        environment_spec: mava_specs.MAEnvironmentSpec,
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        shared_weights: bool = True,
    ):
        super().__init__(
            environment_spec=environment_spec,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            shared_weights=shared_weights,
        )

    def _get_actor_specs(
        self,
    ) -> Dict[str, acme_specs.Array]:
        obs_specs_per_type: Dict[str, acme_specs.Array] = {}

        agents_by_type = self._env_spec.get_agents_by_type()

        for agent_type, agents in agents_by_type.items():
            actor_state_shape = self._env_spec.get_extra_specs()["s_t"].shape
            obs_specs_per_type[agent_type] = tf.TensorSpec(
                shape=actor_state_shape,
                dtype=tf.dtypes.float32,
            )

        actor_obs_specs = {}
        for agent_key in self._actor_agent_keys:
            agent_type = agent_key.split("_")[0]
            # Get observation spec for actor.
            actor_obs_specs[agent_key] = obs_specs_per_type[agent_type]
        return actor_obs_specs


class StateBasedQValueCritic(DecentralisedQValueActorCritic):
    """Multi-agent actor critic architecture with a critic using
    environment state information."""

    def __init__(
        self,
        environment_spec: mava_specs.MAEnvironmentSpec,
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        shared_weights: bool = True,
    ):
        super().__init__(
            environment_spec=environment_spec,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            shared_weights=shared_weights,
        )

    def _get_critic_specs(
        self,
    ) -> Tuple[Dict[str, acme_specs.Array], Dict[str, acme_specs.Array]]:
        action_specs_per_type: Dict[str, acme_specs.Array] = {}

        agents_by_type = self._env_spec.get_agents_by_type()

        # Create one critic per agent. Each critic gets
        # absolute state information of the environment.
        critic_state_shape = self._env_spec.get_extra_specs()["s_t"].shape
        critic_obs_spec = tf.TensorSpec(
            shape=critic_state_shape,
            dtype=tf.dtypes.float32,
        )
        for agent_type, agents in agents_by_type.items():
            critic_act_shape = list(
                copy.copy(self._agent_specs[agents[0]].actions.shape)
            )
            critic_act_shape.insert(0, len(agents))
            action_specs_per_type[agent_type] = tf.TensorSpec(
                shape=critic_act_shape,
                dtype=tf.dtypes.float32,
            )

        critic_obs_specs = {}
        critic_act_specs = {}
        for agent_key in self._critic_agent_keys:
            agent_type = agent_key.split("_")[0]
            # Get observation and action spec for critic.
            critic_obs_specs[agent_key] = critic_obs_spec
            critic_act_specs[agent_key] = action_specs_per_type[agent_type]
        return critic_obs_specs, critic_act_specs


class StateBasedQValueActorCritic(  # type: ignore
    StateBasedPolicyActor, StateBasedQValueCritic
):
    """Multi-agent actor critic architecture where both actor policies
    and critics use environment state information"""

    def __init__(
        self,
        environment_spec: mava_specs.MAEnvironmentSpec,
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        shared_weights: bool = True,
    ):

        StateBasedQValueCritic.__init__(
            self,
            environment_spec=environment_spec,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            shared_weights=shared_weights,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (Claude): complete base class for stabilisation

import abc
from typing import Dict

import sonnet as snt

"""Base stabilising interface for multi-agent RL systems"""


class BaseStabilisationModule:
    """Base class for MARL stabilising.
    Objects which implement this interface provide a set of functions
    to create systems that can stabilise training for agents in a
    multi-agent RL system.
    """

    @abc.abstractmethod
    def some_abstract_stabilising_function(self) -> Dict[str, Dict[str, snt.Module]]:
        """Abstract communication function."""

    @abc.abstractmethod
    def create_system(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create system architecture with stabilisation."""



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Tuple, Union

import gym
import numpy as np

"""
A simple two-step cooperative matrix game for two agents. Adapted from
Qmix paper https://arxiv.org/abs/1803.11485 and tensorflow implementation
https://github.com/tocom242242/qmix_tf2/blob/master/two_step_env.py.

Actions:
0 or 1 - corresponds with which matrix the agents choose for the next time
step.

Observations:
0, 1 or 2 - corresponds with which state the system of agents is in. Both
agents in the system will always have the same state. System starts in
state 0 and moves to state 1 or state 2 depending on the actions of agent
1 in the first timestep. System state resets after both agents act in
timestep 2.
"""


class TwoStepEnv(gym.Env):
    def __init__(self) -> None:
        self.num_agents = 2
        self.reset()
        self.agent_ids = []

        for a_i in range(self.num_agents):
            agent_id = "agent_" + str(a_i)
            self.agent_ids.append(agent_id)

        self.possible_agents = self.agent_ids

    def reset(
        self,
    ) -> Tuple[Dict[str, Union[np.array, Any]], Dict[str, Any]]:
        self.state = 0
        self.reward_n = {
            "agent_0": np.array(0.0, dtype=np.float32),
            "agent_1": np.array(0.0, dtype=np.float32),
        }
        self.env_done = False
        self.done_n = {"agent_0": False, "agent_1": False}
        self.obs_n = {
            "agent_0": np.array([0.0], dtype=np.float32),
            "agent_1": np.array([0.0], dtype=np.float32),
        }
        return self.obs_n, {"s_t": np.array(self.state, dtype=np.int64)}

    def step(
        self, action_n: Dict[str, int]
    ) -> Tuple[
        Dict[str, Union[np.array, Any]],
        Union[dict, Dict[str, Union[float, Any]]],
        Dict[str, Any],
        Dict[str, Any],
    ]:
        if self.state == 0:
            if action_n["agent_0"] == 0:
                self.state = 1
                self.obs_n = {
                    "agent_0": np.array([1.0], dtype=np.float32),
                    "agent_1": np.array([1.0], dtype=np.float32),
                }
                return (
                    self.obs_n,
                    self.reward_n,
                    self.done_n,
                    {"s_t": np.array(self.state, dtype=np.int64)},
                )  # Go to 2A
            else:
                self.state = 2
                self.obs_n = {
                    "agent_0": np.array([2.0], dtype=np.float32),
                    "agent_1": np.array([2.0], dtype=np.float32),
                }
                return (
                    self.obs_n,
                    self.reward_n,
                    self.done_n,
                    {"s_t": np.array(self.state, dtype=np.int64)},
                )  # Go to 2B

        elif self.state == 1:  # State 2A
            self.env_done = True
            self.state = 0
            self.reward_n = {
                "agent_0": np.array(7.0, dtype=np.float32),
                "agent_1": np.array(7.0, dtype=np.float32),
            }
            self.done_n = {"agent_0": True, "agent_1": True}
            return (
                self.obs_n,
                self.reward_n,
                self.done_n,
                {"s_t": np.array(self.state, dtype=np.int64)},
            )

        elif self.state == 2:  # State 2B
            self.env_done = True
            self.state = 0
            self.done_n = {"agent_0": True, "agent_1": True}
            if action_n["agent_0"] == 0 and action_n["agent_1"] == 0:
                self.reward_n = {
                    "agent_0": np.array(0.0, dtype=np.float32),
                    "agent_1": np.array(0.0, dtype=np.float32),
                }
            elif action_n["agent_0"] == 0 and action_n["agent_1"] == 1:
                self.reward_n = {
                    "agent_0": np.array(1.0, dtype=np.float32),
                    "agent_1": np.array(1.0, dtype=np.float32),
                }
            elif action_n["agent_0"] == 1 and action_n["agent_1"] == 0:
                self.reward_n = {
                    "agent_0": np.array(1.0, dtype=np.float32),
                    "agent_1": np.array(1.0, dtype=np.float32),
                }
            elif action_n["agent_0"] == 1 and action_n["agent_1"] == 1:
                self.reward_n = {
                    "agent_0": np.array(8.0, dtype=np.float32),
                    "agent_1": np.array(8.0, dtype=np.float32),
                }
            return (
                self.obs_n,
                self.reward_n,
                self.done_n,
                {"s_t": np.array(self.state, dtype=np.int64)},
            )

        else:
            raise Exception("invalid state:{}".format(self.state))



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Adapted from
# https://github.com/deepmind/acme/blob/master/acme/adders/reverb/episode.py

"""Episode adders.

This implements full episode adders, potentially with padding.
"""

from typing import Optional

import dm_env
import reverb
import tensorflow as tf
import tree
from acme import specs, types
from acme.adders.reverb import utils
from acme.utils import tree_utils

from mava.adders.reverb import base
from mava.adders.reverb import utils as mava_utils


class ParallelEpisodeAdder(base.ReverbParallelAdder):
    """An adder which adds sequences of fixed length."""

    def __init__(
        self,
        client: reverb.Client,
        max_sequence_length: int,
        delta_encoded: bool = False,
        chunk_length: Optional[int] = None,
        priority_fns: Optional[base.PriorityFnMapping] = None,
        pad_end_of_episode: bool = True,
        break_end_of_episode: bool = True,
        max_in_flight_items: Optional[int] = 25,
    ):
        """Makes a SequenceAdder instance.

        Args:
          client: See docstring for BaseAdder.
          max_sequence_length: The maximum length of an episode.
          period: The period with which we add sequences. If less than
            sequence_length, overlapping sequences are added. If equal to
            sequence_length, sequences are exactly non-overlapping.
          delta_encoded: If `True` (False by default) enables delta encoding, see
            `Client` for more information.
          chunk_length: Number of timesteps grouped together before delta encoding
            and compression. See `Client` for more information.
          priority_fns: See docstring for BaseAdder.
          pad_end_of_episode: If True (default) then upon end of episode the current
            sequence will be padded (with observations, actions, etc... whose values
            are 0) until its length is `sequence_length`. If False then the last
            sequence in the episode may have length less than `sequence_length`.
          break_end_of_episode: If 'False' (True by default) does not break
            sequences on env reset. In this case 'pad_end_of_episode' is not used.
          max_in_flight_items: The maximum number of items allowed to be "in flight"
            at the same time. See `reverb.Writer.writer` for more info.
        """

        super().__init__(
            client=client,
            buffer_size=max_sequence_length - 1,
            max_sequence_length=max_sequence_length,
            delta_encoded=delta_encoded,
            chunk_length=chunk_length,
            priority_fns=priority_fns,
            max_in_flight_items=max_in_flight_items,
        )

    def add(
        self,
        action: types.NestedArray,
        next_timestep: dm_env.TimeStep,
        extras: types.NestedArray = (),
    ) -> None:
        if len(self._buffer) == self._buffer.maxlen:
            # If the buffer is full that means we've buffered max_sequence_length-1
            # steps, one dangling observation, and are trying to add one more (which
            # will overflow the buffer).
            raise ValueError(
                "The number of observations within the same episode exceeds "
                "max_sequence_length"
            )

        super().add(action, next_timestep, extras)

    def _write(self) -> None:
        # Append the previous step and increment number of steps written.
        self._writer.append(self._buffer[-1])

    def _write_last(self) -> None:
        # Create a final step.
        # TODO (Dries): Should self._next_observation be used
        #  here? Should this function be used for sequential?
        final_step = mava_utils.final_step_like(
            self._buffer[0], self._next_observations, self._next_extras
        )

        # Append the final step.
        self._writer.append(final_step)

        # The length of the sequence we will be adding is the size of the buffer
        # plus one due to the final step.
        steps = list(self._buffer) + [final_step]
        num_steps = len(steps)

        # Calculate the priority for this episode.
        table_priorities = utils.calculate_priorities(self._priority_fns, steps)

        # Create a prioritized item for each table.
        for table_name, priority in table_priorities.items():
            self._writer.create_item(table_name, num_steps, priority)

    @classmethod
    def signature(
        cls,
        environment_spec: specs.EnvironmentSpec,
        extras_specs: tf.TypeSpec,
    ) -> tf.TypeSpec:
        """This is a helper method for generating signatures for Reverb tables.

        Signatures are useful for validating data types and shapes, see Reverb's
        documentation for details on how they are used.

        Args:
          environment_spec: A `specs.EnvironmentSpec` whose fields are nested
            structures with leaf nodes that have `.shape` and `.dtype` attributes.
            This should come from the environment that will be used to generate
            the data inserted into the Reverb table.
          extras_spec: A nested structure with leaf nodes that have `.shape` and
            `.dtype` attributes. The structure (and shapes/dtypes) of this must
            be the same as the `extras` passed into `ReverbAdder.add`.

        Returns:
          A `Step` whose leaf nodes are `tf.TensorSpec` objects.
        """
        agent_specs = environment_spec.get_agent_specs()
        agents = environment_spec.get_agent_ids()
        extras_specs.update(environment_spec.get_extra_specs())
        obs_specs = {}
        act_specs = {}
        reward_specs = {}
        step_discount_specs = {}
        for agent in agents:
            rewards_spec, step_discounts_spec = tree_utils.broadcast_structures(
                agent_specs[agent].rewards, agent_specs[agent].discounts
            )
            obs_specs[agent] = agent_specs[agent].observations
            act_specs[agent] = agent_specs[agent].actions
            reward_specs[agent] = rewards_spec
            step_discount_specs[agent] = step_discounts_spec

        spec_step = base.Step(
            observations=obs_specs,
            actions=act_specs,
            rewards=reward_specs,
            discounts=step_discount_specs,
            start_of_episode=specs.Array(shape=(), dtype=bool),
            extras=extras_specs,
        )
        return tree.map_structure_with_path(base.spec_like_to_tensor_spec, spec_step)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for adders which transmit data to a replay buffer."""
import abc
from typing import Dict

import dm_env
from acme import adders, types

DEFAULT_PRIORITY_TABLE = "priority_table"


class ParallelAdder(adders.Adder):
    """The Adder interface.
    An adder packs together data to send to the replay buffer, and potentially
    performs some reduction/transformation to this data in the process.
    All adders will use this API. Below is an illustrative example of how they
    are intended to be used in a typical RL run-loop. We assume that the
    environment conforms to the dm_env environment API.
    ```python
    # Reset the environment and add the first observation.
    timestep = env.reset()
    adder.add_first(timestep.observation)
    while not timestep.last():
        # Generate an action from the policy and step the environment.
        action = my_policy(timestep)
        timestep = env.step(action)
        # Add the action and the resulting timestep.
        adder.add(action, next_timestep=timestep)
    ```
    Note that for all adders, the `add()` method expects an action taken and the
    *resulting* timestep observed after taking this action. Note that this
    timestep is named `next_timestep` precisely to emphasize this point.
    """

    @abc.abstractmethod
    def add(
        self,
        actions: Dict[str, types.NestedArray],
        next_timestep: dm_env.TimeStep,
        extras: Dict[str, types.NestedArray] = {"": ()},
    ) -> None:
        """Defines the adder `add` interface.
        Args:
          actions: Dictionary of a possibly nested structure corresponding to
            a_t for each agent.
          next_timestep: A dm_env Timestep object corresponding to the resulting
            data obtained by taking the given action.
          extras: Dictionary of a possibly nested structure of extra data to add
            to replay.
        """



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mava.wrappers.debugging_envs import DebuggingEnvWrapper, TwoStepWrapper
from mava.wrappers.env_wrappers import ParallelEnvWrapper, SequentialEnvWrapper
from mava.wrappers.environment_loop_wrappers import (
    DetailedEpisodeStatistics,
    DetailedPerAgentStatistics,
    MonitorParallelEnvironmentLoop,
)
from mava.wrappers.pettingzoo import (
    PettingZooAECEnvWrapper,
    PettingZooParallelEnvWrapper,
)
from mava.wrappers.robocup import RoboCupWrapper

# TODO only import smac if the user needs it.
from mava.wrappers.smac import SMACEnvWrapper
from mava.wrappers.system_trainer_statistics import (
    DetailedTrainerStatistics,
    DetailedTrainerStatisticsWithEpsilon,
    NetworkStatisticsActorCritic,
)



import os
from typing import Any, Dict, Iterable, Sequence

import sonnet as snt
import tensorflow as tf
import trfl


# Checkpoint the networks.
def checkpoint_networks(system_checkpointer: Dict) -> None:
    if system_checkpointer and len(system_checkpointer.keys()) > 0:
        for network_key in system_checkpointer.keys():
            checkpointer = system_checkpointer[network_key]
            checkpointer.save()


def set_growing_gpu_memory() -> None:
    # Solve gpu mem issues.
    os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "true"
    physical_devices = tf.config.list_physical_devices("GPU")
    if physical_devices:
        for device in physical_devices:
            tf.config.experimental.set_memory_growth(device, True)


# Map critic and polic losses to dict, grouped by agent.
def map_losses_per_agent_ac(critic_losses: Dict, policy_losses: Dict) -> Dict:
    assert len(policy_losses) > 0 and (
        len(critic_losses) == len(policy_losses)
    ), "Invalid System Checkpointer."
    logged_losses: Dict[str, Dict[str, Any]] = {}
    agents = policy_losses.keys()
    for agent in agents:
        logged_losses[agent] = {
            "critic_loss": critic_losses[agent],
            "policy_loss": policy_losses[agent],
        }

    return logged_losses


def combine_dim(tensor: tf.Tensor) -> tf.Tensor:
    dims = tensor.shape[:2]
    return snt.merge_leading_dims(tensor, num_dims=2), dims


def extract_dim(tensor: tf.Tensor, dims: tf.Tensor) -> tf.Tensor:
    return tf.reshape(tensor, [dims[0], dims[1], -1])


# Require correct tensor ranks---as long as we have shape information
# available to check. If there isn't any, we print a warning.
def check_rank(tensors: Iterable[tf.Tensor], ranks: Sequence[int]) -> None:
    for i, (tensor, rank) in enumerate(zip(tensors, ranks)):
        if tensor.get_shape():
            trfl.assert_rank_and_shape_compatibility([tensor], rank)
        else:
            raise ValueError(
                f'Tensor "{tensor.name}", which was offered as '
                f"transformed_n_step_loss parameter {i+1}, has "
                f"no rank at construction time, so cannot verify"
                f"that it has the necessary rank of {rank}"
            )


# Safely delete object from class.
def safe_del(object_class: Any, attrname: str) -> None:
    try:
        if hasattr(object_class, attrname):
            delattr(object_class, attrname)
    except AttributeError:
        pass



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A multi-agent/environment training loop for Petting Zoo."""
from typing import Union

import numpy as np

from mava.wrappers.debugging_envs import DebuggingEnvWrapper


def get_good_simple_spread_action(
    agent_id: int, obs: np.array, environment: DebuggingEnvWrapper
) -> Union[int, np.array]:
    import gym

    diff = np.array(obs[5:7])
    if type(environment.action_spaces[agent_id]) == gym.spaces.discrete.Discrete:
        x, y = diff
        if abs(x) > abs(y):
            if x < 0.0:
                return 1
            else:
                return 2
        else:
            if y < 0.0:
                return 3
            else:
                return 4
    elif type(environment.action_spaces[agent_id]) == gym.spaces.box.Box:
        return diff
    else:
        raise ValueError(
            "Unknown action space: ", type(environment.action_spaces[agent_id])
        )


def test_loop(environment: DebuggingEnvWrapper) -> None:
    while True:
        timestep = environment.reset()
        obs_n = timestep.observation
        tot_rewards = 0
        while not environment.env_done:
            action_n = {}
            for agent_id in environment.agent_ids:
                action_n[agent_id] = get_good_simple_spread_action(
                    agent_id, obs_n[agent_id].observation, environment
                )

            timestep = environment.step(action_n)
            obs_n = timestep.observation
            reward_n = timestep.reward

            tot_rewards += np.mean([r for r in reward_n.values()])
        print("Done. mean total rewards: ", tot_rewards)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for nested data structures involving NumPy and TensorFlow 2.x."""

from typing import List, Optional

import sonnet as snt
import tensorflow as tf
import tree
from acme import types
from acme.tf.utils import add_batch_dim, squeeze_batch_dim, zeros_like

from mava.types import OLT


def ones_like(nest: types.Nest) -> types.NestedTensor:
    """Given a nest of array-like objects, returns similarly nested tf.zeros."""
    return tree.map_structure(lambda x: tf.ones(x.shape, x.dtype), nest)


def create_variables(
    network: snt.Module,
    input_spec: List[OLT],
) -> Optional[tf.TensorSpec]:
    """Builds the network with dummy inputs to create the necessary variables.
    Args:
      network: Sonnet Module whose variables are to be created.
      input_spec: list of input specs to the network. The length of this list
        should match the number of arguments expected by `network`.
    Returns:
      output_spec: only returns an output spec if the output is a tf.Tensor, else
          it doesn't return anything (None); e.g. if the output is a
          tfp.distributions.Distribution.
    """
    # Create a dummy observation with no batch dimension.
    dummy_input = [
        OLT(
            observation=zeros_like(in_spec.observation),
            legal_actions=ones_like(in_spec.legal_actions),
            terminal=zeros_like(in_spec.terminal),
        )
        for in_spec in input_spec
    ]

    # If we have an RNNCore the hidden state will be an additional input.
    if isinstance(network, snt.RNNCore):
        initial_state = squeeze_batch_dim(network.initial_state(1))
        dummy_input += [initial_state]

    # Forward pass of the network which will create variables as a side effect.
    dummy_output = network(*add_batch_dim(dummy_input))

    # Evaluate the input signature by converting the dummy input into a
    # TensorSpec. We then save the signature as a property of the network. This is
    # done so that we can later use it when creating snapshots. We do this here
    # because the snapshot code may not have access to the precise form of the
    # inputs.
    input_signature = tree.map_structure(
        lambda t: tf.TensorSpec((None,) + t.shape, t.dtype), dummy_input
    )
    network._input_signature = input_signature  # pylint: disable=protected-access

    def spec(output: tf.Tensor) -> tf.TensorSpec:
        # If the output is not a Tensor, return None as spec is ill-defined.
        if not isinstance(output, tf.Tensor):
            return None
        # If this is not a scalar Tensor, make sure to squeeze out the batch dim.
        if tf.rank(output) > 0:
            output = squeeze_batch_dim(output)
        return tf.TensorSpec(output.shape, output.dtype)

    return tree.map_structure(spec, dummy_output)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps a Debugging MARL environment to be used as a dm_env environment."""
from typing import Dict, Tuple

import dm_env
import numpy as np
from acme import specs

# from acme.specs import EnvironmentSpec
from acme.wrappers.gym_wrapper import _convert_to_spec
from gym import spaces

from mava.types import OLT
from mava.utils.debugging.environment import MultiAgentEnv
from mava.utils.debugging.environments.switch_game import MultiAgentSwitchGame
from mava.utils.debugging.environments.two_step import TwoStepEnv
from mava.utils.wrapper_utils import convert_np_type, parameterized_restart
from mava.wrappers.pettingzoo import PettingZooParallelEnvWrapper


class DebuggingEnvWrapper(PettingZooParallelEnvWrapper):
    """Environment wrapper for Debugging MARL environments."""

    def __init__(
        self,
        environment: MultiAgentEnv,
        return_state_info: bool = False,
    ):
        super().__init__(environment=environment)

        self.return_state_info = return_state_info

    def reset(self) -> dm_env.TimeStep:
        """Resets the episode."""
        self._reset_next_step = False
        self._step_type = dm_env.StepType.FIRST
        discount_spec = self.discount_spec()
        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self._environment.possible_agents
        }
        observe, env_extras = self._environment.reset()

        observations = self._convert_observations(
            observe, {agent: False for agent in self.possible_agents}
        )
        rewards_spec = self.reward_spec()
        rewards = {
            agent: convert_np_type(rewards_spec[agent].dtype, 0)
            for agent in self.possible_agents
        }
        if not self.return_state_info:
            env_extras = {}

        return parameterized_restart(rewards, self._discounts, observations), env_extras

    def step(self, actions: Dict[str, np.ndarray]) -> Tuple[dm_env.TimeStep, np.array]:
        """Steps the environment."""

        if self._reset_next_step:
            return self.reset()

        observations, rewards, dones, state = self._environment.step(actions)

        rewards_spec = self.reward_spec()
        #  Handle empty rewards
        if not rewards:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, 0)
                for agent in self.agent_ids
            }
        else:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, reward)
                for agent, reward in rewards.items()
            }

        if observations:
            observations = self._convert_observations(observations, dones)

        if self._environment.env_done:
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        timestep = dm_env.TimeStep(
            observation=observations,
            reward=rewards,
            discount=self._discounts,
            step_type=self._step_type,
        )
        if self.return_state_info:
            return timestep, {"s_t": state}
        else:
            return timestep

    # Convert Debugging environment observation so it's dm_env compatible.
    # Also, the list of legal actions must be converted to a legal actions mask.
    def _convert_observations(
        self, observes: Dict[str, np.ndarray], dones: Dict[str, bool]
    ) -> Dict[str, OLT]:
        observations: Dict[str, OLT] = {}
        for agent, observation in observes.items():
            if isinstance(observation, dict) and "action_mask" in observation:
                legals = observation["action_mask"]
                observation = observation["observation"]
            else:
                # TODO Handle legal actions better for continuous envs,
                #  maybe have min and max for each action and clip the agents actions
                #  accordingly
                legals = np.ones(
                    _convert_to_spec(self._environment.action_spaces[agent]).shape,
                    dtype=self._environment.action_spaces[agent].dtype,
                )

            observation = np.array(observation, dtype=np.float32)
            observations[agent] = OLT(
                observation=observation,
                legal_actions=legals,
                terminal=np.asarray([dones[agent]], dtype=np.float32),
            )

        return observations

    def observation_spec(self) -> Dict[str, OLT]:
        observation_specs = {}
        for agent in self._environment.agent_ids:
            observation_specs[agent] = OLT(
                observation=_convert_to_spec(
                    self._environment.observation_spaces[agent]
                ),
                legal_actions=_convert_to_spec(self._environment.action_spaces[agent]),
                terminal=specs.Array((1,), np.float32),
            )
        return observation_specs

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        extras = {}
        if self.return_state_info:
            shape = self.environment._get_state().shape

            ex_spec = specs.BoundedArray(
                shape=shape,
                dtype="float32",
                name="observation",
                minimum=[float("-inf")] * shape[0],
                maximum=[float("inf")] * shape[0],
            )
            extras.update({"s_t": ex_spec})
        return extras


class SwitchGameWrapper(PettingZooParallelEnvWrapper):
    """Environment wrapper for Debugging Switch environment."""

    def __init__(self, environment: MultiAgentSwitchGame):
        super().__init__(environment=environment)

    def step(self, actions: Dict[str, np.ndarray]) -> dm_env.TimeStep:
        if self._reset_next_step:
            return self.reset()

        observations, rewards, dones, infos = self._environment.step(actions)

        if observations:
            observations = self._convert_observations(observations, dones)

        if self._environment.env_done:
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        return dm_env.TimeStep(
            observation=observations,
            reward=rewards,
            discount=self._discounts,
            step_type=self._step_type,
        )

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return {
            # agent: _convert_to_spec(
            #     spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
            # )
            # for agent in self._environment.agent_ids
        }


class TwoStepWrapper(PettingZooParallelEnvWrapper):
    """Wraps simple two-step matrix game from Qmix paper. Useful for
    debugging and quick comparison of cooperative performance."""

    def __init__(self, environment: TwoStepEnv) -> None:
        super().__init__(environment=environment)
        self._reset_next_step = False
        self.environment.action_spaces = {}
        self.environment.observation_spaces = {}
        self.environment.extra_specs = {
            "s_t": spaces.Discrete(3)
        }  # Global state 1, 2, or 3

        for agent_id in self.environment.agent_ids:
            self.environment.action_spaces[agent_id] = spaces.Discrete(2)  # int64
            self.environment.observation_spaces[agent_id] = spaces.Box(
                0, 1, shape=(1,)
            )  # float32

        self.reset()

    def step(self, actions: Dict[str, np.array]) -> Tuple[dm_env.TimeStep, np.array]:
        """Steps the environment."""
        if self._reset_next_step:
            self._reset_next_step = False
            self.reset()

        observations, rewards, dones, state_infos = self._environment.step(actions)
        if observations:
            observations = self._convert_observations(observations, dones)

        if self._environment.env_done:
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        return (
            dm_env.TimeStep(
                observation=observations,
                reward=rewards,
                discount=self._discounts,
                step_type=self._step_type,
            ),
            state_infos,
        )

    def reset(self) -> Tuple[dm_env.TimeStep, np.array]:
        """Resets the episode."""
        self._reset_next_step = False
        self._step_type = dm_env.StepType.FIRST
        discount_spec = self.discount_spec()
        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self._environment.possible_agents
        }
        observe, state_infos = self._environment.reset()
        observations = self._convert_observations(
            observe, {agent: False for agent in self.possible_agents}
        )
        rewards_spec = self.reward_spec()
        rewards = {
            agent: convert_np_type(rewards_spec[agent].dtype, 0)
            for agent in self.possible_agents
        }

        discount_spec = self.discount_spec()
        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self.possible_agents
        }
        return (
            parameterized_restart(rewards, self._discounts, observations),
            state_infos,
        )

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return self.environment.extra_specs



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Any, Dict, Iterator, List, Optional, Type, Union

import numpy as np
import reverb
import sonnet as snt
from acme import datasets
from acme.tf import variable_utils
from acme.utils import counting

from mava import adders, core, specs, types
from mava.adders import reverb as reverb_adders
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)
from mava.components.tf.modules.stabilising import FingerPrintStabalisation
from mava.systems.tf import executors
from mava.systems.tf.madqn import execution, training
from mava.wrappers import DetailedTrainerStatisticsWithEpsilon


@dataclasses.dataclass
class MADQNConfig:
    """Configuration options for the MADDPG system.
    Args:
            environment_spec: description of the actions, observations, etc.
            discount: discount to use for TD updates.
            batch_size: batch size for updates.
            prefetch_size: size to prefetch from replay.
            target_update_period: number of learner steps to perform before updating
              the target networks.
            min_replay_size: minimum replay size before updating.
            max_replay_size: maximum replay size.
            samples_per_insert: number of samples to take from replay for every insert
              that is made.
            n_step: number of steps to squash into a single transition.
            sigma: standard deviation of zero-mean, Gaussian exploration noise.
            clipping: whether to clip gradients by global norm.
            replay_table_name: string indicating what name to give the replay table."""

    environment_spec: specs.MAEnvironmentSpec
    epsilon_min: float
    epsilon_decay: float
    shared_weights: bool
    max_gradient_norm: Optional[float]
    target_update_period: int
    executor_variable_update_period: int
    min_replay_size: int
    max_replay_size: int
    samples_per_insert: Optional[float]
    prefetch_size: int
    batch_size: int
    n_step: int
    sequence_length: int
    period: int
    discount: float
    checkpoint: bool
    optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]]
    replay_table_name: str = reverb_adders.DEFAULT_PRIORITY_TABLE
    checkpoint_subpath: str = "~/mava/"


class MADQNBuilder:
    """Builder for MADQN which constructs individual components of the system."""

    """Defines an interface for defining the components of an RL system.
      Implementations of this interface contain a complete specification of a
      concrete RL system. An instance of this class can be used to build an
      RL system which interacts with the environment either locally or in a
      distributed setup.
      """

    def __init__(
        self,
        config: MADQNConfig,
        trainer_fn: Type[training.MADQNTrainer] = training.MADQNTrainer,
        executor_fn: Type[core.Executor] = execution.MADQNFeedForwardExecutor,
        extra_specs: Dict[str, Any] = {},
        exploration_scheduler_fn: Type[
            LinearExplorationScheduler
        ] = LinearExplorationScheduler,
        replay_stabilisation_fn: Optional[Type[FingerPrintStabalisation]] = None,
    ):
        """Args:
        _config: Configuration options for the MADQN system.
        _trainer_fn: Trainer module to use."""
        self._config = config
        self._extra_specs = extra_specs

        self._agents = self._config.environment_spec.get_agent_ids()
        self._agent_types = self._config.environment_spec.get_agent_types()
        self._trainer_fn = trainer_fn
        self._executor_fn = executor_fn
        self._exploration_scheduler_fn = exploration_scheduler_fn
        self._replay_stabiliser_fn = replay_stabilisation_fn

    def make_replay_tables(
        self,
        environment_spec: specs.MAEnvironmentSpec,
    ) -> List[reverb.Table]:
        """Create tables to insert data into."""

        # Select adder
        if issubclass(self._executor_fn, executors.FeedForwardExecutor):
            # Check if we should use fingerprints
            if self._replay_stabiliser_fn is not None:
                self._extra_specs.update({"fingerprint": np.array([1.0, 1.0])})
            adder_sig = reverb_adders.ParallelNStepTransitionAdder.signature(
                environment_spec, self._extra_specs
            )
        elif issubclass(self._executor_fn, executors.RecurrentExecutor):
            adder_sig = reverb_adders.ParallelSequenceAdder.signature(
                environment_spec, self._extra_specs
            )
        else:
            raise NotImplementedError("Unknown executor type: ", self._executor_fn)

        if self._config.samples_per_insert is None:
            # We will take a samples_per_insert ratio of None to mean that there is
            # no limit, i.e. this only implies a min size limit.
            limiter = reverb.rate_limiters.MinSize(self._config.min_replay_size)

        else:
            # Create enough of an error buffer to give a 10% tolerance in rate.
            samples_per_insert_tolerance = 0.1 * self._config.samples_per_insert
            error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
            limiter = reverb.rate_limiters.SampleToInsertRatio(
                min_size_to_sample=self._config.min_replay_size,
                samples_per_insert=self._config.samples_per_insert,
                error_buffer=error_buffer,
            )

        replay_table = reverb.Table(
            name=self._config.replay_table_name,
            sampler=reverb.selectors.Uniform(),
            remover=reverb.selectors.Fifo(),
            max_size=self._config.max_replay_size,
            rate_limiter=limiter,
            signature=adder_sig,
        )

        return [replay_table]

    def make_dataset_iterator(
        self, replay_client: reverb.Client
    ) -> Iterator[reverb.ReplaySample]:
        """Create a dataset iterator to use for learning/updating the system.
        Args:
            replay_client: Reverb Client which points to the replay server."""

        sequence_length = (
            self._config.sequence_length
            if issubclass(self._executor_fn, executors.RecurrentExecutor)
            else None
        )

        dataset = datasets.make_reverb_dataset(
            table=self._config.replay_table_name,
            server_address=replay_client.server_address,
            batch_size=self._config.batch_size,
            prefetch_size=self._config.prefetch_size,
            sequence_length=sequence_length,
        )
        return iter(dataset)

    def make_adder(
        self, replay_client: reverb.Client
    ) -> Optional[adders.ParallelAdder]:
        """Create an adder which records data generated by the executor/environment.
        Args:
          replay_client: Reverb Client which points to the replay server."""

        # Select adder
        if issubclass(self._executor_fn, executors.FeedForwardExecutor):
            adder = reverb_adders.ParallelNStepTransitionAdder(
                priority_fns=None,
                client=replay_client,
                n_step=self._config.n_step,
                discount=self._config.discount,
            )
        elif issubclass(self._executor_fn, executors.RecurrentExecutor):
            adder = reverb_adders.ParallelSequenceAdder(
                priority_fns=None,
                client=replay_client,
                sequence_length=self._config.sequence_length,
                period=self._config.period,
            )
        else:
            raise NotImplementedError("Unknown executor type: ", self._executor_fn)

        return adder

    def make_executor(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, Any],
        adder: Optional[adders.ParallelAdder] = None,
        variable_source: Optional[core.VariableSource] = None,
        trainer: Optional[training.MADQNTrainer] = None,
        communication_module: Optional[BaseCommunicationModule] = None,
        evaluator: bool = False,
    ) -> core.Executor:
        """Create an executor instance.
        Args:
            q_networks: A struct of instance of all
                the different system q networks,
                this should be a callable which takes as input observations
                and returns actions.
            adder: How data is recorded (e.g. added to replay).
            variable_source: collection of (nested) numpy arrays. Contains
                source variables as defined in mava.core.
        """

        shared_weights = self._config.shared_weights

        variable_client = None
        if variable_source:
            agent_keys = self._agent_types if shared_weights else self._agents

            # Create policy variables
            variables = {agent: q_networks[agent].variables for agent in agent_keys}
            # Get new policy variables
            variable_client = variable_utils.VariableClient(
                client=variable_source,
                variables={"q_network": variables},
                update_period=self._config.executor_variable_update_period,
            )

            # Make sure not to use a random policy after checkpoint restoration by
            # assigning variables before running the environment loop.
            variable_client.update_and_wait()

        # Check if we should use fingerprints
        fingerprint = True if self._replay_stabiliser_fn is not None else False

        # Create the executor which coordinates the actors.
        return self._executor_fn(
            q_networks=q_networks,
            action_selectors=action_selectors,
            shared_weights=shared_weights,
            variable_client=variable_client,
            adder=adder,
            trainer=trainer,
            communication_module=communication_module,
            evaluator=evaluator,
            fingerprint=fingerprint,
        )

    def make_trainer(
        self,
        networks: Dict[str, Dict[str, snt.Module]],
        dataset: Iterator[reverb.ReplaySample],
        counter: Optional[counting.Counter] = None,
        logger: Optional[types.NestedLogger] = None,
        communication_module: Optional[BaseCommunicationModule] = None,
    ) -> core.Trainer:
        """Creates an instance of the trainer.
        Args:
          networks: struct describing the networks needed by the trainer; this can
            be specific to the trainer in question.
          dataset: iterator over samples from replay.
          counter: a Counter which allows for recording of counts (trainer steps,
            executor steps, etc.) distributed throughout the system.
          logger: Logger object for logging metadata.
          checkpoint: bool controlling whether the trainer checkpoints itself.
        """
        q_networks = networks["values"]
        target_q_networks = networks["target_values"]

        agents = self._config.environment_spec.get_agent_ids()
        agent_types = self._config.environment_spec.get_agent_types()

        # Make epsilon scheduler
        exploration_scheduler = self._exploration_scheduler_fn(
            epsilon_min=self._config.epsilon_min,
            epsilon_decay=self._config.epsilon_decay,
        )

        # Check if we should use fingerprints
        fingerprint = True if self._replay_stabiliser_fn is not None else False

        # The learner updates the parameters (and initializes them).
        trainer = self._trainer_fn(
            agents=agents,
            agent_types=agent_types,
            discount=self._config.discount,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            shared_weights=self._config.shared_weights,
            optimizer=self._config.optimizer,
            target_update_period=self._config.target_update_period,
            max_gradient_norm=self._config.max_gradient_norm,
            exploration_scheduler=exploration_scheduler,
            communication_module=communication_module,
            dataset=dataset,
            counter=counter,
            fingerprint=fingerprint,
            logger=logger,
            checkpoint=self._config.checkpoint,
            checkpoint_subpath=self._config.checkpoint_subpath,
        )

        trainer = DetailedTrainerStatisticsWithEpsilon(trainer)  # type:ignore

        return trainer



# Adapted from https://github.com/openai/multiagent-particle-envs.
# TODO (dries): Try using this class directly from PettingZoo and delete this file.

# An old version of OpenAI Gym's multi_discrete.py.
# (Was getting affected by Gym updates)
# (https://github.com/openai/gym/blob/1fb81d4e3fb780ccf77fec731287ba07da35eb84
# /gym/spaces/multi_discrete.py)

from typing import List, Tuple

import gym
import numpy as np


class MultiDiscrete(gym.Space):
    """
    - The multi-discrete action space consists of a series
    of discrete action spaces with different parameters
    - It can be adapted to both a Discrete action space or
    a continuous (Box) action space
    - It is useful to represent game controllers or keyboards
    where each key can be represented as a discrete action space
    - It is parametrized by passing an array of arrays containing
    [min, max] for each discrete action space
       where the discrete action space can take any integers
       from `min` to `max` (both inclusive)
    Note: A value of 0 always need to represent the NOOP action.
    e.g. Nintendo Game Controller
    - Can be conceptualized as 3 discrete action spaces:
        1) Arrow Keys: Discrete 5  - NOOP[0], UP[1], RIGHT[2],
        DOWN[3], LEFT[4]  - params: min: 0, max: 4
        2) Button A:   Discrete 2  - NOOP[0], Pressed[1] -
        params: min: 0, max: 1
        3) Button B:   Discrete 2  - NOOP[0], Pressed[1] -
        params: min: 0, max: 1
    - Can be initialized as
        MultiDiscrete([ [0,4], [0,1], [0,1] ])
    """

    def __init__(self, array_of_param_array: np.array) -> None:
        self.low = np.array([x[0] for x in array_of_param_array])
        self.high = np.array([x[1] for x in array_of_param_array])
        self.num_discrete_space = self.low.shape[0]

    def sample(self) -> List[int]:
        """Returns a array with one sample from each discrete action space.
        For each row:  round(random .* (max - min) + min, 0)"""
        random_array = np.random.RandomState().rand(self.num_discrete_space)
        return [
            int(x)
            for x in np.floor(
                np.multiply((self.high - self.low + 1.0), random_array) + self.low
            )
        ]

    def contains(self, x: List) -> Tuple[bool]:
        return (
            len(x) == self.num_discrete_space
            and (np.array(x) >= self.low).all()
            and (np.array(x) <= self.high).all()
        )

    @property
    def shape(self) -> Tuple:
        return self.num_discrete_space

    def __repr__(self) -> str:
        return "MultiDiscrete" + str(self.num_discrete_space)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Dict

import sonnet as snt

"""Base architecture interface for multi-agent RL systems"""


class BaseArchitecture:
    """Base class for MARL architectures.
    Objects which implement this interface provide a set of functions
    to create systems according to a specific architectural design,
    e.g. decentralised, centralised or networked.
    """

    @abc.abstractmethod
    def create_actor_variables(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create network variables for actors in the system."""

    @abc.abstractmethod
    def create_system(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create system architecture."""


class BasePolicyArchitecture(BaseArchitecture):
    """Base class for policy gradient MARL architectures.
    Objects which implement this interface provide a set of functions
    to create systems according to a specific architectural design,
    e.g. decentralised, centralised or networked.
    """

    @abc.abstractmethod
    def create_behaviour_policy(self) -> Dict[str, Dict[str, snt.Module]]:
        """Return behaviour policy networks (observation network + policy head)."""


class BaseActorCritic(BasePolicyArchitecture):
    """Base class for MARL Actor critic architectures.
    Objects which implement this interface provide a set of functions
    to create systems according to a specific architectural design,
    e.g. decentralised, centralised or networked.
    """

    @abc.abstractmethod
    def create_critic_variables(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create network variables for critics in the system."""



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# See SMAC here: https://github.com/oxwhirl/smac
# Documentation available at smac/blob/master/docs/smac.md

"""Wraps a StarCraft II MARL environment (SMAC) as a dm_env environment."""

from typing import Any, Dict, List, Tuple, Type

import dm_env
import numpy as np
from acme import specs
from acme.wrappers.gym_wrapper import _convert_to_spec
from gym.spaces import Box, Discrete
from pettingzoo.utils.env import ParallelEnv

# TODO should not be imported unless user wants the smac env.
from smac.env import StarCraft2Env  # type:ignore

from mava import types
from mava.utils.wrapper_utils import convert_np_type, parameterized_restart
from mava.wrappers.env_wrappers import ParallelEnvWrapper


class SMACEnvWrapper(ParallelEnvWrapper):
    """
    Wraps a StarCraft II MARL environment (SMAC) as a Mava Parallel environment.
    Based on RLlib wrapper provided by SMAC.
    """

    def __init__(self, environment: StarCraft2Env) -> None:
        """Create a new multi-agent StarCraft env compatible with RLlib.
        Arguments:
            smac_args (dict): Arguments to pass to the underlying
                smac.env.starcraft.StarCraft2Env instance.
        """
        self._environment = environment

        self._reset_next_step = True
        self._possible_agents = [
            f"agent_{i}" for i in range(self._environment.get_env_info()["n_agents"])
        ]
        self._agents = self._possible_agents
        self.observation_space = {
            "observation": Box(-1, 1, shape=(self._environment.get_obs_size(),)),
            "action_mask": Box(0, 1, shape=(self._environment.get_total_actions(),)),
        }
        self.observation_spaces = {
            agent: self.observation_space["observation"] for agent in self._agents
        }
        self.action_space: Type[Discrete] = Discrete(
            self._environment.get_total_actions()
        )
        self.action_spaces = {agent: self.action_space for agent in self._agents}

    def reset(self) -> Tuple[dm_env.TimeStep, np.array]:
        """Resets the env and returns observations from ready agents.
        Returns:
            obs (dict): New observations for each ready agent.
        """
        self._env_done = False
        self._reset_next_step = False
        self._step_type = dm_env.StepType.FIRST

        # reset internal SC2 env
        obs_list, state = self._environment.reset()

        # Convert observations
        observe: Dict[str, np.ndarray] = {}

        for i, obs in enumerate(obs_list):
            observe[f"agent_{i}"] = {
                "observation": obs,
                "action_mask": np.array(
                    self._environment.get_avail_agent_actions(i), dtype=np.float32
                ),
            }

        observations = self._convert_observations(
            observe, {agent: False for agent in self._possible_agents}
        )

        self._agents = list(observe.keys())

        # create discount spec
        discount_spec = self.discount_spec()
        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self._possible_agents
        }

        # create rewards spec
        rewards_spec = self.reward_spec()
        rewards = {
            agent: convert_np_type(rewards_spec[agent].dtype, 0)
            for agent in self._possible_agents
        }

        # dm_env timestep
        timestep = parameterized_restart(rewards, self._discounts, observations)

        return timestep, {"s_t": state}

    def step(self, actions: Dict[str, np.ndarray]) -> Tuple[dm_env.TimeStep, np.array]:
        """Returns observations from ready agents.
        The returns are dicts mapping from agent_id strings to values. The
        number of agents in the env can vary over time.
        Returns
        -------
            obs (dict): New observations for each ready agent.
            rewards (dict): Reward values for each ready agent. If the
                episode is just started, the value will be None.
            dones (dict): Done values for each ready agent. The special key
                "__all__" (required) is used to indicate env termination.
            infos (dict): Optional info values for each agent id.
        """
        if self._reset_next_step:
            return self.reset()

        actions_feed = list(actions.values())
        reward, terminated, info = self._environment.step(actions_feed)
        obs_list = self._environment.get_obs()
        state = self._environment.get_state()
        self._env_done = terminated

        observe = {}
        rewards = {}
        dones = {}
        for i, obs in enumerate(obs_list):
            agent = f"agent_{i}"
            observe[agent] = {
                "observation": obs,
                "action_mask": np.array(
                    self._environment.get_avail_agent_actions(i), dtype=np.float32
                ),
            }
            rewards[agent] = reward
            dones[agent] = terminated

        observations = self._convert_observations(observe, dones)
        self._agents = list(observe.keys())
        rewards_spec = self.reward_spec()

        #  Handle empty rewards
        if not rewards:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, 0)
                for agent in self._possible_agents
            }
        else:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, reward)
                for agent, reward in rewards.items()
            }

        if self.env_done():
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        timestep = dm_env.TimeStep(
            observation=observations,
            reward=rewards,
            discount=self._discounts,
            step_type=self._step_type,
        )

        return timestep, {"s_t": state}

    def env_done(self) -> bool:
        """
        Returns a bool indicating if all agents in env are done.
        """
        return self._env_done

    def _convert_observations(
        self, observes: Dict[str, np.ndarray], dones: Dict[str, bool]
    ) -> types.Observation:
        observations: Dict[str, types.OLT] = {}
        for agent, observation in observes.items():
            if isinstance(observation, dict) and "action_mask" in observation:
                legals = observation["action_mask"]
                observation = observation["observation"]
            else:
                legals = np.ones(
                    _convert_to_spec(self.action_space).shape,
                    dtype=self.action_space.dtype,
                )
            observations[agent] = types.OLT(
                observation=observation,
                legal_actions=legals,
                terminal=np.asarray([dones[agent]], dtype=np.float32),
            )

        return observations

    def observation_spec(self) -> types.Observation:
        return {
            agent: types.OLT(
                observation=_convert_to_spec(self.observation_space["observation"]),
                legal_actions=_convert_to_spec(self.observation_space["action_mask"]),
                terminal=specs.Array((1,), np.float32),
            )
            for agent in self._possible_agents
        }

    def action_spec(self) -> Dict[str, specs.DiscreteArray]:
        return {
            agent: _convert_to_spec(self.action_space)
            for agent in self._possible_agents
        }

    def reward_spec(self) -> Dict[str, specs.Array]:
        return {agent: specs.Array((), np.float32) for agent in self._possible_agents}

    def discount_spec(self) -> Dict[str, specs.BoundedArray]:
        return {
            agent: specs.BoundedArray((), np.float32, minimum=0, maximum=1.0)
            for agent in self._possible_agents
        }

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        state = self._environment.get_state()
        # TODO (dries): What should the real bounds be of the state spec?
        return {
            "s_t": specs.BoundedArray(
                state.shape, np.float32, minimum=float("-inf"), maximum=float("inf")
            )
        }

    def seed(self, random_seed: int) -> None:
        self._environment._seed = random_seed

    @property
    def agents(self) -> List:
        return self._agents

    @property
    def possible_agents(self) -> List:
        return self._possible_agents

    @property
    def environment(self) -> ParallelEnv:
        """Returns the wrapped environment."""
        return self._environment

    def __getattr__(self, name: str) -> Any:
        """Expose any other attributes of the underlying environment."""
        return getattr(self._environment, name)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (Claude): complete class for importance sampling stabilisation

"""Stabilising for multi-agent RL systems"""

from mava.components.tf.architectures import BaseArchitecture
from mava.components.tf.modules.stabilising import BaseStabilisationModule


class ImportanceSamplingStabalisation(BaseStabilisationModule):
    """Multi-agent stabalisation architecture."""

    def __init__(
        self,
        architecture: BaseArchitecture,
    ) -> None:
        self._architecture = architecture

    def some_stabilising_function(self) -> None:
        """Perform some stabilising logic"""



# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tree
from acme import types
from acme.adders.reverb import utils as acme_utils

from mava.adders.reverb import base


def final_step_like(
    step: base.Step, next_observations: types.NestedArray, next_extras: dict = None
) -> base.Step:
    """Return a list of steps with the final step zero-filled."""
    # Make zero-filled components so we can fill out the last step.
    zero_action, zero_reward, zero_discount = tree.map_structure(
        acme_utils.zeros_like, (step.actions, step.rewards, step.discounts)
    )

    return base.Step(
        observations=next_observations,
        actions=zero_action,
        rewards=zero_reward,
        discounts=zero_discount,
        start_of_episode=False,
        extras=next_extras
        if next_extras
        else tree.map_structure(acme_utils.zeros_like, step.extras),
    )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sonnet module that takes two inputs
[observation, fingerprint]"""

# import launchpad as lp
import sonnet as snt
import tensorflow as tf


class ObservationNetworkWithFingerprint(snt.Module):
    """Sonnet module that takes two inputs
    [observation, fingerprint] and returns an observation
    embedding and concatenates the fingerprint to the
    embedding. Downstream layers can then be trained
    on the embedding+fingerprint."""

    def __init__(
        self,
        observation_network: snt.Module,
    ) -> None:
        """Initializes network.
        Args:
            observation_network: ...
        """
        super(ObservationNetworkWithFingerprint, self).__init__()
        self._observation_network = observation_network
        self._flatten_layer = tf.keras.layers.Flatten()

    def __call__(
        self,
        obs: tf.Tensor,
        fingerprint: tf.Tensor,
    ) -> tf.Tensor:

        hidden = self._observation_network(obs)
        flatten = self._flatten_layer(hidden)
        hidden_with_fingerprint = tf.concat([flatten, fingerprint], axis=1)

        return hidden_with_fingerprint



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import typing
from typing import Any, Dict, List, Tuple, Union

import acme
import dm_env
import numpy as np
import numpy.testing as npt
import pytest
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.rail_generators import sparse_rail_generator
from flatland.envs.schedule_generators import sparse_schedule_generator
from pettingzoo.utils.env import AECEnv, ParallelEnv

from mava import specs as mava_specs
from mava.environment_loop import ParallelEnvironmentLoop, SequentialEnvironmentLoop
from mava.types import Observation, Reward
from mava.utils.environments.flatland_utils import load_flatland_env
from mava.utils.wrapper_utils import convert_np_type
from mava.wrappers.flatland import FlatlandEnvWrapper
from mava.wrappers.pettingzoo import (
    PettingZooAECEnvWrapper,
    PettingZooParallelEnvWrapper,
)
from tests.enums import EnvSource, EnvSpec, EnvType, MockedEnvironments
from tests.mocks import (
    ParallelMAContinuousEnvironment,
    ParallelMADiscreteEnvironment,
    SequentialMAContinuousEnvironment,
    SequentialMADiscreteEnvironment,
)

# flatland environment config
rail_gen_cfg: Dict = {
    "max_num_cities": 4,
    "max_rails_between_cities": 2,
    "max_rails_in_city": 3,
    "grid_mode": True,
    "seed": 42,
}

flatland_env_config: Dict = {
    "number_of_agents": 2,
    "width": 25,
    "height": 25,
    "rail_generator": sparse_rail_generator(**rail_gen_cfg),
    "schedule_generator": sparse_schedule_generator(),
    "obs_builder_object": TreeObsForRailEnv(max_depth=2),
}


"""
Helpers contains re-usable test functions.
"""

# TODO(Kale-ab): Better structure helper funcs


class Helpers:
    # Check all props are not none
    @staticmethod
    def verify_all_props_not_none(props_which_should_not_be_none: list) -> bool:
        return all(prop is not None for prop in props_which_should_not_be_none)

    # Return an env - currently Pettingzoo envs.
    @staticmethod
    def get_env(env_spec: EnvSpec) -> Union[AECEnv, ParallelEnv]:
        env = None
        if env_spec.env_source == EnvSource.PettingZoo:
            mod = importlib.import_module(env_spec.env_name)
            if env_spec.env_type == EnvType.Parallel:
                env = mod.parallel_env()  # type:ignore
            elif env_spec.env_type == EnvType.Sequential:
                env = mod.env()  # type:ignore
        elif env_spec.env_source == EnvSource.Flatland:
            env = load_flatland_env(flatland_env_config)
        else:
            raise Exception("Env_spec is not valid.")
        env.reset()  # type:ignore
        return env

    # Returns a wrapper function.
    @staticmethod
    def get_wrapper_function(
        env_spec: EnvSpec,
    ) -> dm_env.Environment:
        wrapper: dm_env.Environment = None
        if env_spec.env_source == EnvSource.PettingZoo:
            if env_spec.env_type == EnvType.Parallel:
                wrapper = PettingZooParallelEnvWrapper
            elif env_spec.env_type == EnvType.Sequential:
                wrapper = PettingZooAECEnvWrapper
        elif env_spec.env_source == EnvSource.Flatland:
            wrapper = FlatlandEnvWrapper
        else:
            raise Exception("Env_spec is not valid.")
        return wrapper

    # Returns an env loop.
    @staticmethod
    def get_env_loop(
        env_spec: EnvSpec,
    ) -> acme.core.Worker:
        env_loop = None
        if env_spec.env_type == EnvType.Parallel:
            env_loop = ParallelEnvironmentLoop
        elif env_spec.env_type == EnvType.Sequential:
            env_loop = SequentialEnvironmentLoop
        else:
            raise Exception("Env_spec is not valid.")
        return env_loop

    """Function that retrieves a mocked env, based on
    env_spec."""

    @staticmethod
    def get_mocked_env(
        env_spec: EnvSpec,
    ) -> Union[
        ParallelMADiscreteEnvironment,
        SequentialMADiscreteEnvironment,
        ParallelMAContinuousEnvironment,
        SequentialMAContinuousEnvironment,
    ]:
        env = None
        if not hasattr(env_spec, "env_name"):
            raise Exception("No env_name passed in.")

        if not hasattr(env_spec, "env_type"):
            raise Exception("No env_type passed in.")
        env_name = env_spec.env_name

        if env_name is MockedEnvironments.Mocked_Dicrete:
            if env_spec.env_type == EnvType.Parallel:
                env = ParallelMADiscreteEnvironment(
                    num_actions=18,
                    num_observations=2,
                    obs_shape=(84, 84, 4),
                    obs_dtype=np.float32,
                    episode_length=10,
                )
            elif env_spec.env_type == EnvType.Sequential:
                env = SequentialMADiscreteEnvironment(
                    num_actions=18,
                    num_observations=2,
                    obs_shape=(84, 84, 4),
                    obs_dtype=np.float32,
                    episode_length=10,
                )
        elif env_name is MockedEnvironments.Mocked_Continous:
            if env_spec.env_type == EnvType.Parallel:
                env = ParallelMAContinuousEnvironment(
                    action_dim=2,
                    observation_dim=2,
                    bounded=True,
                    episode_length=10,
                )
            elif env_spec.env_type == EnvType.Sequential:
                env = SequentialMAContinuousEnvironment(
                    action_dim=2,
                    observation_dim=2,
                    bounded=True,
                    episode_length=10,
                )

        if env is None:
            raise Exception("Env_spec is not valid.")
        return env

    @staticmethod
    def is_mocked_env(
        env_name: str,
    ) -> bool:
        mock = False
        if (
            env_name is MockedEnvironments.Mocked_Continous
            or env_name is MockedEnvironments.Mocked_Dicrete
        ):
            mock = True
        return mock

    # Returns a wrapped env and specs
    @staticmethod
    def get_wrapped_env(
        env_spec: EnvSpec, **kwargs: Any
    ) -> Tuple[dm_env.Environment, acme.specs.EnvironmentSpec]:

        specs = None
        if Helpers.is_mocked_env(env_spec.env_name):
            wrapped_env = Helpers.get_mocked_env(env_spec)
            specs = wrapped_env._specs
        else:
            env = Helpers.get_env(env_spec)
            wrapper_func = Helpers.get_wrapper_function(env_spec)
            wrapped_env = wrapper_func(env, **kwargs)
            specs = Helpers.get_pz_env_spec(wrapped_env)._specs
        wrapped_env.reset()  # type : ignore
        return wrapped_env, specs

    # Returns a petting zoo environment spec.
    @staticmethod
    def get_pz_env_spec(environment: dm_env.Environment) -> dm_env.Environment:
        return mava_specs.MAEnvironmentSpec(environment)

    # Seeds action space
    @staticmethod
    def seed_action_space(
        env_wrapper: Union[PettingZooAECEnvWrapper, PettingZooParallelEnvWrapper],
        random_seed: int,
    ) -> None:
        [
            env_wrapper.action_spaces[agent].seed(random_seed)
            for agent in env_wrapper.agents
        ]

    @staticmethod
    def compare_dicts(dictA: Dict, dictB: Dict) -> bool:
        typesA = [type(k) for k in dictA.values()]
        typesB = [type(k) for k in dictB.values()]

        return (dictA == dictB) and (typesA == typesB)

    @staticmethod
    def assert_valid_episode(episode_result: Dict) -> None:
        assert (
            episode_result["episode_length"] > 0
            and episode_result["mean_episode_return"] is not None
            and episode_result["steps_per_second"] is not None
            and episode_result["episodes"] == 1
            and episode_result["steps"] > 0
        )

    @staticmethod
    def assert_env_reset(
        wrapped_env: dm_env.Environment,
        dm_env_timestep: dm_env.TimeStep,
        env_spec: EnvSpec,
    ) -> None:
        if env_spec.env_type == EnvType.Parallel:
            rewards_spec = wrapped_env.reward_spec()
            expected_rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, 0)
                for agent in wrapped_env.agents
            }

            discount_spec = wrapped_env.discount_spec()
            expected_discounts = {
                agent: convert_np_type(rewards_spec[agent].dtype, 1)
                for agent in wrapped_env.agents
            }

            Helpers.compare_dicts(
                dm_env_timestep.reward,
                expected_rewards,
            ), "Failed to reset reward."
            Helpers.compare_dicts(
                dm_env_timestep.discount,
                expected_discounts,
            ), "Failed to reset discount."

        elif env_spec.env_type == EnvType.Sequential:
            for agent in wrapped_env.agents:
                rewards_spec = wrapped_env.reward_spec()
                expected_reward = convert_np_type(rewards_spec[agent].dtype, 0)

                discount_spec = wrapped_env.discount_spec()
                expected_discount = convert_np_type(discount_spec[agent].dtype, 1)

                assert dm_env_timestep.reward == expected_reward and type(
                    dm_env_timestep.reward
                ) == type(expected_reward), "Failed to reset reward."
                assert dm_env_timestep.discount == expected_discount and type(
                    dm_env_timestep.discount
                ) == type(expected_discount), "Failed to reset discount."

    @staticmethod
    @typing.no_type_check
    # TODO(Kale-ab) Sort out typing issues.
    def verify_observations_are_normalized(
        observations: Observation,
        agents: List,
        env_spec: EnvSpec,
        min: int = 0,
        max: int = 1,
    ) -> None:
        if env_spec.env_type == EnvType.Parallel:
            for agent in agents:
                assert (
                    observations[agent].observation.min() >= min
                    and observations[agent].observation.max() <= max
                ), "Failed to normalize observations."

        elif env_spec.env_type == EnvType.Sequential:
            assert (
                observations.observation.min() >= min
                and observations.observation.max() <= max
            ), "Failed to normalize observations."

    @staticmethod
    @typing.no_type_check
    # TODO(Kale-ab) Sort out typing issues.
    def verify_reward_is_normalized(
        rewards: Reward, agents: List, env_spec: EnvSpec, min: int = 0, max: int = 1
    ) -> None:
        if env_spec.env_type == EnvType.Parallel:
            for agent in agents:
                assert (
                    rewards[agent] >= min and rewards[agent] <= max
                ), "Failed to normalize reward."

        elif env_spec.env_type == EnvType.Sequential:
            assert rewards >= min and rewards <= max, "Failed to normalize reward."

    @staticmethod
    def verify_observations_are_standardized(
        observations: Observation, agents: List, env_spec: EnvSpec
    ) -> None:
        if env_spec.env_type == EnvType.Parallel:
            for agent in agents:
                npt.assert_almost_equal(
                    observations[agent].observation.mean(), 0, decimal=2
                )
                npt.assert_almost_equal(
                    observations[agent].observation.std(), 1, decimal=2
                )

        elif env_spec.env_type == EnvType.Sequential:
            npt.assert_almost_equal(
                observations.observation.mean(), 0, decimal=2  # type: ignore
            )
            npt.assert_almost_equal(
                observations.observation.std(), 1, decimal=2  # type: ignore
            )

    @staticmethod
    def mock_done() -> bool:
        return True


@typing.no_type_check
@pytest.fixture
def helpers() -> Helpers:
    return Helpers



# Adapted from https://github.com/openai/multiagent-particle-envs.

from typing import List

import numpy as np

from mava.utils.debugging.core import Agent, Landmark, World
from mava.utils.debugging.scenario import BaseScenario


class Scenario(BaseScenario):
    def make_world(self, num_agents: int) -> World:
        world = World()
        # set any world properties first
        num_landmarks = num_agents
        # add agents
        world.agents = [Agent() for i in range(num_agents)]
        for i, agent in enumerate(world.agents):
            agent.name = "agent %d" % i
            agent.collide = True
            agent.size = 0.15
        # add landmarks
        world.landmarks = [Landmark() for i in range(num_landmarks)]
        for i, landmark in enumerate(world.landmarks):
            landmark.name = "landmark %d" % i
            landmark.collide = False
            landmark.movable = False
        # make initial conditions
        self.reset_world(world)
        return world

    def reset_world(self, world: World) -> None:
        # random properties for agents
        for i, agent in enumerate(world.agents):
            agent.color = np.array([0.35, 0.35, 0.85])
        # random properties for landmarks
        for i, landmark in enumerate(world.landmarks):
            landmark.color = np.array([0.25, 0.25, 0.25])
        # set random initial states
        for agent in world.agents:
            agent.state.p_pos = np.random.uniform(-1, +1, world.dim_p)
            agent.state.p_vel = np.zeros(world.dim_p)
        for i, landmark in enumerate(world.landmarks):
            landmark.state.p_pos = np.random.uniform(-1, +1, world.dim_p)
            landmark.state.p_vel = np.zeros(world.dim_p)

        # Reset step counter
        world.current_step = 0

    def is_collision(self, agent1: Agent, agent2: Agent) -> bool:
        delta_pos = agent1.state.p_pos - agent2.state.p_pos
        dist = np.sqrt(np.sum(np.square(delta_pos)))
        dist_min = agent1.size + agent2.size
        return True if dist < dist_min else False

    @staticmethod
    def dist(pt1: np.array, pt2: np.array) -> float:
        return np.sqrt(np.sum(np.square(pt1 - pt2)))

    def reward(self, agent: Agent, a_i: int, world: World) -> float:
        # Agents are rewarded based on agent distance to its corresponding
        # landmark, penalized for collisions
        rew = 0.0
        # for i in range(len(world.agents)):
        other_agent = world.agents[a_i]
        landmark = world.landmarks[a_i]
        distance = self.dist(other_agent.state.p_pos, landmark.state.p_pos)

        # Cap distance
        if distance > 1.0:
            distance = 1.0
        rew += 1 - distance

        if agent.collide:
            for other in world.agents:
                if other is agent:
                    continue
                if self.is_collision(other, agent):
                    rew -= 1

        return rew

    def observation(self, agent: Agent, a_i: int, world: World) -> np.array:
        # get positions of all entities in this agent's reference frame
        entity_pos = []
        for entity in world.landmarks:  # world.entities:
            entity_pos.append(entity.state.p_pos - agent.state.p_pos)
        # entity colors
        entity_color: List[np.array] = []
        for entity in world.landmarks:  # world.entities:
            entity_color.append(entity.color)
        other_pos = []
        other_landmarks = []
        for i, other in enumerate(world.agents):
            if other is agent:
                continue
            landmark = world.landmarks[i]
            other_landmarks.append(landmark.state.p_pos - landmark.state.p_pos)
            other_pos.append(other.state.p_pos - agent.state.p_pos)

        return np.concatenate(
            [agent.state.p_vel]
            + [agent.state.p_pos]
            + [[world.current_step / 50]]
            + [entity_pos[a_i]]
            + other_pos
            + other_landmarks  # + comm
        )

    def done(self, agent: Agent, world: World) -> bool:
        if world.current_step < 50:
            return False
        else:
            return True



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MAD4PG system implementation."""
from typing import Any, Dict, Type, Union

from mava import core
from mava.systems.tf import executors
from mava.systems.tf.mad4pg import training
from mava.systems.tf.maddpg.builder import MADDPGBuilder, MADDPGConfig


class MAD4PGBuilder(MADDPGBuilder):
    """Builder for MAD4PG which constructs individual components of the system."""

    """Defines an interface for defining the components of an RL system.
      Implementations of this interface contain a complete specification of a
      concrete RL system. An instance of this class can be used to build an
      RL system which interacts with the environment either locally or in a
      distributed setup.
      """

    def __init__(
        self,
        config: MADDPGConfig,
        trainer_fn: Union[
            Type[training.MAD4PGBaseTrainer],
            Type[training.MAD4PGBaseRecurrentTrainer],
        ] = training.MAD4PGDecentralisedTrainer,
        executor_fn: Type[core.Executor] = executors.FeedForwardExecutor,
        extra_specs: Dict[str, Any] = {},
    ):
        """Args:
        config: Configuration options for the MAD4PG system.
        trainer_fn: Trainer module to use."""

        super().__init__(
            config=config,
            trainer_fn=trainer_fn,
            executor_fn=executor_fn,
            extra_specs=extra_specs,
        )



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MAD4PG system implementation."""
from typing import Callable, Dict, Optional, Type, Union

import dm_env
import sonnet as snt
from acme import specs as acme_specs

from mava import core
from mava import specs as mava_specs
from mava.components.tf.architectures import DecentralisedQValueActorCritic
from mava.environment_loop import ParallelEnvironmentLoop
from mava.systems.tf.mad4pg import training
from mava.systems.tf.maddpg.execution import MADDPGFeedForwardExecutor
from mava.systems.tf.maddpg.system import MADDPG
from mava.utils.loggers import MavaLogger


class MAD4PG(MADDPG):
    """MAD4PG system.
    This implements a single-process D4PG system. This is an actor-critic based
    system that generates data via a behavior policy, inserts N-step transitions into
    a replay buffer, and periodically updates the policies of each agent
    (and as a result the behavior) by sampling uniformly from this buffer.
    """

    def __init__(
        self,
        environment_factory: Callable[[bool], dm_env.Environment],
        network_factory: Callable[[acme_specs.BoundedArray], Dict[str, snt.Module]],
        logger_factory: Callable[[str], MavaLogger] = None,
        architecture: Type[
            DecentralisedQValueActorCritic
        ] = DecentralisedQValueActorCritic,
        trainer_fn: Union[
            Type[training.MAD4PGBaseTrainer],
            Type[training.MAD4PGBaseRecurrentTrainer],
        ] = training.MAD4PGDecentralisedTrainer,
        executor_fn: Type[core.Executor] = MADDPGFeedForwardExecutor,
        num_executors: int = 1,
        num_caches: int = 0,
        environment_spec: mava_specs.MAEnvironmentSpec = None,
        shared_weights: bool = True,
        discount: float = 0.99,
        batch_size: int = 256,
        prefetch_size: int = 4,
        target_averaging: bool = False,
        target_update_period: int = 100,
        target_update_rate: Optional[float] = None,
        executor_variable_update_period: int = 1000,
        min_replay_size: int = 1000,
        max_replay_size: int = 1000000,
        samples_per_insert: float = 32.0,
        policy_optimizer: Union[
            snt.Optimizer, Dict[str, snt.Optimizer]
        ] = snt.optimizers.Adam(learning_rate=1e-4),
        critic_optimizer: snt.Optimizer = snt.optimizers.Adam(learning_rate=1e-4),
        n_step: int = 5,
        sequence_length: int = 20,
        period: int = 20,
        sigma: float = 0.3,
        max_gradient_norm: float = None,
        max_executor_steps: int = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
        logger_config: Dict = {},
        train_loop_fn: Callable = ParallelEnvironmentLoop,
        eval_loop_fn: Callable = ParallelEnvironmentLoop,
        train_loop_fn_kwargs: Dict = {},
        eval_loop_fn_kwargs: Dict = {},
    ):
        """Initialize the system.
        Args:
            environment_factory: Callable to instantiate an environment
                on a compute node.
            network_factory: Callable to instantiate system networks on a compute node.
            logger_factory: Callable to instantiate a system logger on a compute node.
            architecture: system architecture, e.g. decentralised or centralised.
            trainer_fn: training type associated with executor and architecture,
                e.g. centralised training.
            executor_fn: executor type for example feedforward or recurrent.
            num_executors: number of executor processes to run in parallel.
            num_caches: number of trainer node caches.
            environment_spec: description of the actions, observations, etc.
            shared_weights: set whether agents should share network weights.
            discount: discount to use for TD updates.
            batch_size: batch size for updates.
            prefetch_size: size to prefetch from replay.
            target_update_period: number of learner steps to perform before updating
              the target networks.
            min_replay_size: minimum replay size before updating.
            max_replay_size: maximum replay size.
            samples_per_insert: number of samples to take from replay for every insert
              that is made.
            n_step: number of steps to squash into a single transition.
            sequence_length: Length of the sequences to use in recurrent
            training (if using recurrence).
            period: Overlapping period of sequences used in recurrent
            training (if using recurrence).
            sigma: standard deviation of zero-mean, Gaussian exploration noise.
            max_gradient_norm:
            logger: logger object to be used by trainers.
            counter: counter object used to keep track of steps.
            checkpoint: boolean indicating whether to checkpoint the trainers.
            checkpoint_subpath: directory for checkpoints.
            replay_table_name: string indicating what name to give the replay table.
            train_loop_fn: loop for training.
            eval_loop_fn: loop for evaluation.
            policy_optimizer: the optimizer to be applied to the policy loss.
                This can be a single optimizer or an optimizer per agent key.
            critic_optimizer: the optimizer to be applied to the critic loss.
        """
        super().__init__(
            environment_factory=environment_factory,
            network_factory=network_factory,
            logger_factory=logger_factory,
            architecture=architecture,
            trainer_fn=trainer_fn,
            executor_fn=executor_fn,
            num_executors=num_executors,
            num_caches=num_caches,
            environment_spec=environment_spec,
            shared_weights=shared_weights,
            discount=discount,
            batch_size=batch_size,
            prefetch_size=prefetch_size,
            target_update_period=target_update_period,
            executor_variable_update_period=executor_variable_update_period,
            min_replay_size=min_replay_size,
            max_replay_size=max_replay_size,
            samples_per_insert=samples_per_insert,
            policy_optimizer=policy_optimizer,
            critic_optimizer=critic_optimizer,
            n_step=n_step,
            sequence_length=sequence_length,
            period=period,
            sigma=sigma,
            max_gradient_norm=max_gradient_norm,
            max_executor_steps=max_executor_steps,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            logger_config=logger_config,
            train_loop_fn=train_loop_fn,
            eval_loop_fn=eval_loop_fn,
            train_loop_fn_kwargs=train_loop_fn_kwargs,
            eval_loop_fn_kwargs=eval_loop_fn_kwargs,
            target_averaging=target_averaging,
            target_update_rate=target_update_rate,
        )
